import numpy as np
import re
import math
import json
import logging

from qcodes.utils.helpers import deep_update, NumpyJSONEncoder
from .data_array import DataArray
from .format import Formatter
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .data_set import DataSet


log = logging.getLogger(__name__)


class GNUPlotFormat(Formatter):
    """
    Saves data in one or more gnuplot-format files. We make one file for
    each set of matching dependent variables in the loop.

    Args:

        extension (str): file extension for data files. Defaults to
            'dat'

        terminator (str): newline character(s) to use on write
            not used for reading, we will read any combination of '\\\\r'
            and '\\\\n'. Defaults to '\\\\n'

        separator (str): field (column) separator, must be whitespace.
            Only used for writing, we will read with any whitespace separation.
            Defaults to '\\\\t'.

        comment (str): lines starting with this are not data
            Comments are written with this full string, and identified on read
            by just the string after stripping whitespace. Defaults to '# '.

        number_format (str): from the format mini-language, how to
            format numeric data into a string. Defaults to 'g'.

        always_nest (bool): whether to always make a folder for files
            or just make a single data file if all data has the same setpoints.
            Defaults to bool.

    These files are basically tab-separated values, but any quantity of
    any whitespace characters is accepted.

    Each row represents one setting of the setpoint variable(s)
    the setpoint variable(s) are in the first column(s)
    measured variable(s) come after.

    The data is preceded by comment lines (starting with #).
    We use three:

    - one for the variable name
    - the (longer) axis label, in quotes so a label can contain whitespace.
    - for each dependent var, the (max) number of points in that dimension
      (this also tells us how many dependent vars we have in this file)

    ::

        # id1\tid2\tid3...
        # "label1"\t"label2"\t"label3"...
        # 100\t250
        1\t2\t3...
        2\t3\t4...

    For data of 2 dependent variables, gnuplot puts each inner loop into one
    block, then increments the outer loop in the next block, separated by a
    blank line.

    We extend this to an arbitrary quantity of dependent variables by using
    one blank line for each loop level that resets. (gnuplot *does* seem to
    use 2 blank lines sometimes, to denote a whole new dataset, which sort
    of corresponds to our situation.)
    """

    def __init__(self, extension='dat', terminator='\n', separator='\t',
                 comment='# ', number_format='.15g', metadata_file=None):
        self.metadata_file = metadata_file or 'snapshot.json'
        # file extension: accept either with or without leading dot
        self.extension = '.' + extension.lstrip('.')

        # line terminator (only used for writing; will read any \r\n combo)
        if terminator not in ('\r', '\n', '\r\n'):
            raise ValueError(
                r'GNUPlotFormat terminator must be \r, \n, or \r\n')
        self.terminator = terminator

        # field separator (only used for writing; will read any whitespace)
        if not re.fullmatch(r'\s+', separator):
            raise ValueError('GNUPlotFormat separator must be whitespace')
        self.separator = separator

        # beginning of a comment line. (when reading, just checks the
        # non-whitespace character(s) of comment
        self.comment = comment
        self.comment_chars = comment.rstrip()
        if not self.comment_chars:
            raise ValueError('comment must have some non-whitespace')
        self.comment_len = len(self.comment_chars)

        # number format (only used for writing; will read any number)
        self.number_format = '{:' + number_format + '}'

    def read_one_file(self, data_set, f, ids_read):
        """
        Called by Formatter.read to bring one data file into
        a DataSet. Setpoint data may be duplicated across multiple files,
        but each measured DataArray must only map to one file.

        args:
            data_set: the DataSet we are reading into
            f: a file-like object to read from
            ids_read: a `set` of array_ids that we have already read.
                when you read an array, check that it's not in this set (except
                setpoints, which can be in several files with different inner loop)
                then add it to the set so other files know not to read it again
        """
        if not f.name.endswith(self.extension):
            return

        arrays = data_set.arrays
        ids = self._read_comment_line(f).split()
        labels = self._get_labels(self._read_comment_line(f))
        shape = tuple(map(int, self._read_comment_line(f).split()))
        ndim = len(shape)

        set_arrays = ()
        data_arrays = []
        indexed_ids = list(enumerate(ids))

        for i, array_id in indexed_ids[:ndim]:
            snap = data_set.get_array_metadata(array_id)

            # setpoint arrays
            set_shape = shape[: i + 1]
            if array_id in arrays:
                set_array = arrays[array_id]
                if set_array.shape != set_shape:
                    raise ValueError(
                        'shapes do not match for set array: ' + array_id)
                if array_id not in ids_read:
                    # it's OK for setpoints to be duplicated across
                    # multiple files, but we should only empty the
                    # array out the first time we see it, so subsequent
                    # reads can check for consistency
                    set_array.clear()
            else:
                set_array = DataArray(label=labels[i], array_id=array_id,
                                      set_arrays=set_arrays, shape=set_shape,
                                      is_setpoint=True, snapshot=snap)
                set_array.init_data()
                data_set.add_array(set_array)

            set_arrays = set_arrays + (set_array, )
            ids_read.add(array_id)

        for i, array_id in indexed_ids[ndim:]:
            snap = data_set.get_array_metadata(array_id)

            # data arrays
            if array_id in ids_read:
                raise ValueError('duplicate data id found: ' + array_id)

            if array_id in arrays:
                data_array = arrays[array_id]
                data_array.clear()
            else:
                data_array = DataArray(label=labels[i], array_id=array_id,
                                       set_arrays=set_arrays, shape=shape,
                                       snapshot=snap)
                data_array.init_data()
                data_set.add_array(data_array)
            data_arrays.append(data_array)
            ids_read.add(array_id)

        indices = [0] * ndim
        first_point = True
        resetting = 0
        for line in f:
            if self._is_comment(line):
                continue

            # ignore leading or trailing whitespace (including in blank lines)
            line = line.strip()

            if not line:
                # each consecutive blank line implies one more loop to reset
                # when we read the next data point. Don't depend on the number
                # of setpoints that change, as there could be weird cases, like
                # bidirectional sweeps, or highly diagonal sweeps, where this
                # is incorrect. Anyway this really only matters for >2D sweeps.
                if not first_point:
                    resetting += 1
                continue

            values = tuple(map(float, line.split()))

            if resetting:
                indices[-resetting - 1] += 1
                indices[-resetting:] = [0] * resetting
                resetting = 0

            for value, set_array in zip(values[:ndim], set_arrays):
                nparray = set_array.ndarray
                myindices = tuple(indices[:nparray.ndim])
                stored_value = nparray[myindices]
                if math.isnan(stored_value):
                    nparray[myindices] = value
                elif stored_value != value:
                    raise ValueError('inconsistent setpoint values',
                                     stored_value, value, set_array.name,
                                     myindices, indices)

            for value, data_array in zip(values[ndim:], data_arrays):
                # set .ndarray directly to avoid the overhead of __setitem__
                # which updates modified_range on every call
                data_array.ndarray[tuple(indices)] = value

            indices[-1] += 1
            first_point = False

        # Since we skipped __setitem__, back up to the last read point and
        # mark it as saved that far.
        # Using mark_saved is better than directly setting last_saved_index
        # because it also ensures modified_range is set correctly.
        indices[-1] -= 1
        for array in set_arrays + tuple(data_arrays):
            array.mark_saved(array.flat_index(indices[:array.ndim]))

    def _is_comment(self, line):
        return line[:self.comment_len] == self.comment_chars

    def _read_comment_line(self, f):
        s = f.readline()
        if not self._is_comment(s):
            raise ValueError('expected a comment line, found:\n' + s)
        return s[self.comment_len:]

    def _get_labels(self, labelstr):
        labelstr = labelstr.strip()
        if labelstr[0] != '"' or labelstr[-1] != '"':
            # fields are *not* quoted
            return labelstr.split()
        else:
            # fields *are* quoted (and escaped)
            parts = re.split('"\\s+"', labelstr[1:-1])
            return [l.replace('\\"', '"').replace('\\\\', '\\') for l in parts]

    # this signature is unfortunatly incompatible with the super class
    # so we have to ignore type errors
    def write(self,  # type: ignore[override]
              data_set: 'DataSet',
              io_manager, location, force_write=False,
              write_metadata=True, only_complete=True,
              filename=None):
        """
        Write updates in this DataSet to storage.

        Will choose append if possible, overwrite if not.

        Args:
            data_set: the data we're storing
            io_manager (io_manager): the base location to write to
            location (str): the file location within io_manager
            only_complete (bool): passed to match_save_range, answers the
                following question: Should we write all available new data,
                or only complete rows? Is used to make sure that everything
                gets written when the DataSet is finalised, even if some
                dataarrays are strange (like, full of nans)
            filename (Optional[str]): Filename to save to. Will override
                the usual naming scheme and possibly overwrite files, so
                use with care. The file will be saved in the normal location.
        """
        arrays = data_set.arrays

        # puts everything with same dimensions together
        groups = self.group_arrays(arrays)
        existing_files = set(io_manager.list(location))
        written_files = set()

        # Every group gets its own datafile
        for group in groups:
            log.debug('Attempting to write the following '
                      'group: {}'.format(group.name))
            # it might be useful to output the whole group as below but it is
            # very verbose
            #log.debug('containing {}'.format(group))

            if filename:
                fn = io_manager.join(location, filename + self.extension)
            else:
                fn = io_manager.join(location, group.name + self.extension)

            written_files.add(fn)

            # fn may or may not be an absolute path depending on the location manager
            # used however, io_manager always returns relative paths so make sure both are
            # relative by calling to_location
            file_exists = io_manager.to_location(fn) in existing_files
            save_range = self.match_save_range(group, file_exists,
                                               only_complete=only_complete)

            if save_range is None:
                log.debug('Cannot match save range, skipping this group.')
                continue

            overwrite = save_range[0] == 0 or force_write
            open_mode = 'w' if overwrite else 'a'
            shape = group.set_arrays[-1].shape

            with io_manager.open(fn, open_mode) as f:
                if overwrite:
                    f.write(self._make_header(group))
                    log.debug('Wrote header to file')

                for i in range(save_range[0], save_range[1] + 1):
                    indices = np.unravel_index(i, shape)

                    # insert a blank line for each loop that reset (to index 0)
                    # note that if *all* indices are zero (the first point)
                    # we won't put any blanks
                    for j, index in enumerate(reversed(indices)):
                        if index != 0:
                            if j:
                                f.write(self.terminator * j)
                            break

                    one_point = self._data_point(group, indices)
                    f.write(self.separator.join(one_point) + self.terminator)
                log.debug('Wrote to file from '
                          '{} to {}'.format(save_range[0], save_range[1]+1))
            # now that we've saved the data, mark it as such in the data.
            # we mark the data arrays and the inner setpoint array. Outer
            # setpoint arrays have different dimension (so would need a
            # different unraveled index) but more importantly could have
            # a different saved range anyway depending on whether there
            # is outer data taken before or after the inner loop. Anyway we
            # never look at the outer setpoint last_saved_index or
            # modified_range, we just assume it's got the values we need.
            for array in group.data + (group.set_arrays[-1],):
                array.mark_saved(save_range[1])

        if write_metadata:
            self.write_metadata(
                data_set, io_manager=io_manager, location=location)

    def write_metadata(self, data_set: 'DataSet', io_manager, location,
                       read_first=True, **kwargs):
        """
        Write all metadata in this DataSet to storage.

        Args:
            data_set: the data we're storing

            io_manager (io_manager): the base location to write to

            location (str): the file location within io_manager

            read_first (Optional[bool]): read previously saved metadata before
                writing? The current metadata will still be the used if
                there are changes, but if the saved metadata has information
                not present in the current metadata, it will be retained.
                Default True.
        """
        if read_first:
            # In case the saved file has more metadata than we have here,
            # read it in first. But any changes to the in-memory copy should
            # override the saved file data.
            memory_metadata = data_set.metadata
            data_set.metadata = {}
            self.read_metadata(data_set)
            deep_update(data_set.metadata, memory_metadata)

        fn = io_manager.join(location, self.metadata_file)
        with io_manager.open(fn, 'w', encoding='utf8') as snap_file:
            json.dump(data_set.metadata, snap_file, sort_keys=False,
                      indent=4, ensure_ascii=False, cls=NumpyJSONEncoder)

    def read_metadata(self, data_set):
        io_manager = data_set.io
        location = data_set.location
        fn = io_manager.join(location, self.metadata_file)
        if io_manager.list(fn):
            with io_manager.open(fn, 'r') as snap_file:
                metadata = json.load(snap_file)
            data_set.metadata.update(metadata)

    def _make_header(self, group):
        ids, labels = [], []
        for array in group.set_arrays + group.data:
            ids.append(array.array_id)
            label = getattr(array, 'label', array.array_id)
            label = label.replace('\\', '\\\\').replace('"', '\\"')
            labels.append('"' + label + '"')

        shape = [str(size) for size in group.set_arrays[-1].shape]
        if len(shape) != len(group.set_arrays):
            raise ValueError('array dimensionality does not match setpoints')

        out = (self._comment_line(ids) + self._comment_line(labels) +
               self._comment_line(shape))

        return out

    def _comment_line(self, items):
        return self.comment + self.separator.join(items) + self.terminator

    def _data_point(self, group, indices):
        for array in group.set_arrays:
            yield self.number_format.format(array[indices[:array.ndim]])

        for array in group.data:
            yield self.number_format.format(array[indices])



import logging

import numpy as np
from numpy.testing import assert_array_almost_equal

from qcodes.instrument.parameter import Parameter


def test_setting_non_gettable_parameter_with_finite_step(caplog):
    initial_value = 0
    step_size = 0.1
    set_value = 1.0

    # when the parameter is initially set from
    # the initial_value the starting point is unknown
    # so this should cause a warning but the parameter should still be set
    with caplog.at_level(logging.WARNING):
        x = Parameter('x', initial_value=initial_value,
                      step=step_size,
                      set_cmd=None)
        assert len(caplog.records) == 1
        assert f"cannot sweep x from None to {initial_value}" in str(caplog.records[0])
    assert x.cache.get() == 0

    # afterwards the stepping should work as expected.
    with caplog.at_level(logging.WARNING):
        caplog.clear()
        assert_array_almost_equal(np.array(x.get_ramp_values(set_value, step_size)),
                                  (np.arange(initial_value+step_size, set_value+step_size, step_size)))
        x.set(set_value)
        assert x.cache.get() == set_value
        assert len(caplog.records) == 0



"""Actions, mainly to be executed in measurement Loops."""
import time

from qcodes.utils.helpers import is_function
from qcodes.utils.threading import thread_map

_NO_SNAPSHOT = {'type': None, 'description': 'Action without snapshot'}


# exception when threading is attempted used to simultaneously
# query the same instrument for several values
class UnsafeThreadingException(Exception):
    pass


def _actions_snapshot(actions, update):
    """Make a list of snapshots from a list of actions."""
    snapshot = []
    for action in actions:
        if hasattr(action, 'snapshot'):
            snapshot.append(action.snapshot(update=update))
        else:
            snapshot.append(_NO_SNAPSHOT)
    return snapshot


class Task:
    """
    A predefined task to be executed within a measurement Loop.

    The first argument should be a callable, to which any subsequent
    args and kwargs (which are evaluated before the loop starts) are passed.

    The args and kwargs are first evaluated if they are found to be callable.

    Keyword Args passed when the Task is called are ignored,
    but are accepted for compatibility with other things happening in a Loop.

    Args:
        func (Callable): Function to executed
        *args: pass to func, after evaluation if callable
        **kwargs: pass to func, after evaluation if callable

    """
    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __call__(self, **ignore_kwargs):
        # If any of the arguments are callable, evaluate them first
        eval_args = [arg() if callable(arg) else arg for arg in self.args]
        eval_kwargs = {k: (v() if callable(v) else v) for k, v in self.kwargs.items()}

        self.func(*eval_args, **eval_kwargs)

    def snapshot(self, update=False):
        """
        Snapshots  task
        Args:
            update (bool): TODO not in use

        Returns:
            dict: snapshot
        """
        return {'type': 'Task', 'func': repr(self.func)}


class Wait:
    """
    A simple class to tell a Loop to wait <delay> seconds.

    This is transformed into a Task within the Loop, such that
    it can do other things (monitor, check for halt) during the delay.

    But for use outside of a Loop, it is also callable (then it just sleeps)

    Args:
        delay: seconds to delay

    Raises:
        ValueError: if delay is negative
    """
    def __init__(self, delay):
        if not delay >= 0:
            raise ValueError(f"delay must be > 0, not {repr(delay)}")
        self.delay = delay

    def __call__(self):
        if self.delay:
            time.sleep(self.delay)

    def snapshot(self, update=False):
        """
        Snapshots  delay
        Args:
            update (bool): TODO not in use

        Returns:
            dict: snapshot
        """
        return {'type': 'Wait', 'delay': self.delay}


class _Measure:
    """
    A callable collection of parameters to measure.

    This should not be constructed manually, only by an ActiveLoop.
    """
    def __init__(self, params_indices, data_set, use_threads):
        self.use_threads = use_threads and len(params_indices) > 1
        # the applicable DataSet.store function
        self.store = data_set.store

        # for performance, pre-calculate which params return data for
        # multiple arrays, and the name mappings
        self.getters = []
        self.param_ids = []
        self.composite = []
        paramcheck = []  # list to check if parameters are unique
        for param, action_indices in params_indices:
            self.getters.append(param.get)

            if param._instrument:
                paramcheck.append((param, param._instrument))

            if hasattr(param, 'names'):
                part_ids = []
                for i in range(len(param.names)):
                    param_id = data_set.action_id_map[action_indices + (i,)]
                    part_ids.append(param_id)
                self.param_ids.append(None)
                self.composite.append(part_ids)
            else:
                param_id = data_set.action_id_map[action_indices]
                self.param_ids.append(param_id)
                self.composite.append(False)

        if self.use_threads:
            insts = [p[1] for p in paramcheck]
            if (len(set(insts)) != len(insts)):
                duplicates = [p for p in paramcheck if insts.count(p[1]) > 1]
                raise UnsafeThreadingException('Can not use threading to '
                                               'read '
                                               'several things from the same '
                                               'instrument. Specifically, you '
                                               'asked for'
                                               ' {}.'.format(duplicates))

    def __call__(self, loop_indices, **ignore_kwargs):
        out_dict = {}
        if self.use_threads:
            out = thread_map(self.getters)
        else:
            out = [g() for g in self.getters]

        for param_out, param_id, composite in zip(out, self.param_ids,
                                                  self.composite):
            if composite:
                for val, part_id in zip(param_out, composite):
                    out_dict[part_id] = val
            else:
                out_dict[param_id] = param_out

        self.store(loop_indices, out_dict)


class _Nest:

    """
    Wrapper to make a callable nested ActiveLoop.

    This should not be constructed manually, only by an ActiveLoop.
    """

    def __init__(self, inner_loop, action_indices):
        self.inner_loop = inner_loop
        self.action_indices = action_indices

    def __call__(self, **kwargs):
        self.inner_loop._run_loop(action_indices=self.action_indices, **kwargs)


class BreakIf:

    """
    Loop action that breaks out of the loop if a condition is truthy.

    Args:
        condition (Callable): a callable taking no arguments.
            Can be a simple function that returns truthy when it's time to quit
    Raises:
        TypeError: if condition is not a callable with no aguments.

    Examples:
            >>> BreakIf(lambda: gates.chan1.get() >= 3)
    """

    def __init__(self, condition):
        if not is_function(condition, 0):
            raise TypeError('BreakIf condition must be a callable with '
                            'no arguments')
        self.condition = condition

    def __call__(self, **ignore_kwargs):
        if self.condition():
            raise _QcodesBreak

    def snapshot(self, update=False):
        """
        Snapshots breakif action
        Args:
            update (bool): TODO not in use

        Returns:
            dict: snapshot

        """
        return {'type': 'BreakIf', 'condition': repr(self.condition)}


class _QcodesBreak(Exception):
    pass



import os
from shutil import copy
from unittest import TestCase

import h5py
import numpy as np

import qcodes.data
from qcodes.data.data_array import DataArray
from qcodes.data.data_set import DataSet, load_data, new_data
from qcodes.data.hdf5_format import HDF5Format, str_to_bool
from qcodes.data.location import FormatLocation
from qcodes.loops import Loop
from qcodes.station import Station
from qcodes.tests.instrument_mocks import MockParabola
from qcodes.utils.helpers import compare_dictionaries

from .data_mocks import DataSet1D, DataSet2D


class TestHDF5_Format(TestCase):
    def setUp(self):
        self.io = DataSet.default_io
        self.formatter = HDF5Format()
        # Set up the location provider to always store test data in
        # "qc.tests.unittest_data
        cur_fp = os.path.dirname(__file__)
        base_fp = os.path.abspath(os.path.join(cur_fp, '../unittest_data'))
        self.loc_provider = FormatLocation(
            fmt=base_fp+'/{date}/#{counter}_{name}_{time}')
        DataSet.location_provider = self.loc_provider

    def checkArraysEqual(self, a, b):
        """
        Checks if arrays are equal
        """
        # Modified from GNUplot would be better to have this in some module
        self.checkArrayAttrs(a, b)
        np.testing.assert_array_equal(a, b)
        if len(a.set_arrays) > 1:
            for i, set_arr in enumerate(a.set_arrays):
                np.testing.assert_array_equal(set_arr, b.set_arrays[i])
        else:
            np.testing.assert_array_equal(a.set_arrays, b.set_arrays)

        for sa, sb in zip(a.set_arrays, b.set_arrays):
            self.checkArrayAttrs(sa, sb)

    def checkArrayAttrs(self, a, b):
        self.assertEqual(a.tolist(), b.tolist())
        self.assertEqual(a.label, b.label)
        self.assertEqual(a.array_id, b.array_id)

    def test_full_write_read_1D(self):
        """
        Test writing and reading a file back in
        """
        # location = self.locations[0]
        data = DataSet1D(name='test1D_full_write',
                         location=self.loc_provider)
        # print('Data location:', os.path.abspath(data.location))
        self.formatter.write(data)
        # Used because the formatter has no nice find file method

        # Test reading the same file through the DataSet.read
        data2 = DataSet(location=data.location, formatter=self.formatter)
        data2.read()
        self.checkArraysEqual(data2.x_set, data.x_set)
        self.checkArraysEqual(data2.y, data.y)
        self.formatter.close_file(data)
        self.formatter.close_file(data2)

    def test_full_write_read_2D(self):
        """
        Test writing and reading a file back in
        """
        data = DataSet2D(location=self.loc_provider, name='test2D')
        self.formatter.write(data)
        # Test reading the same file through the DataSet.read
        data2 = DataSet(location=data.location, formatter=self.formatter)
        data2.read()
        self.checkArraysEqual(data2.x_set, data.x_set)
        self.checkArraysEqual(data2.y_set, data.y_set)
        self.checkArraysEqual(data2.z, data.z)

        self.formatter.close_file(data)
        self.formatter.close_file(data2)

    def test_incremental_write(self):
        data = DataSet1D(location=self.loc_provider, name='test_incremental')
        location = data.location
        data_copy = DataSet1D(False)

        # # empty the data and mark it as unmodified
        data.x_set[:] = float('nan')
        data.y[:] = float('nan')
        data.x_set.modified_range = None
        data.y.modified_range = None

        # simulate writing after every value comes in, even within
        # one row (x comes first, it's the setpoint)
        for i, (x, y) in enumerate(zip(data_copy.x_set, data_copy.y)):
            data.x_set[i] = x
            self.formatter.write(data)
            data.y[i] = y
            self.formatter.write(data)
        data2 = DataSet(location=location, formatter=self.formatter)
        data2.read()
        self.checkArraysEqual(data2.arrays['x_set'], data_copy.arrays['x_set'])
        self.checkArraysEqual(data2.arrays['y'], data_copy.arrays['y'])

        self.formatter.close_file(data)
        self.formatter.close_file(data2)

    def test_metadata_write_read(self):
        """
        Test is based on the snapshot of the 1D dataset.
        Having a more complex snapshot in the metadata would be a better test.
        """
        data = DataSet1D(location=self.loc_provider, name='test_metadata')
        data.snapshot()  # gets the snapshot, not added upon init
        self.formatter.write(data)  # write_metadata is included in write
        data2 = DataSet(location=data.location, formatter=self.formatter)
        data2.read()
        self.formatter.close_file(data)
        self.formatter.close_file(data2)
        metadata_equal, err_msg = compare_dictionaries(
            data.metadata, data2.metadata,
            'original_metadata', 'loaded_metadata')
        self.assertTrue(metadata_equal, msg='\n'+err_msg)

    def test_loop_writing(self):
        # pass
        station = Station()
        MockPar = MockParabola(name='Loop_writing_test')
        station.add_component(MockPar)
        # # added to station to test snapshot at a later stage
        loop = Loop(MockPar.x[-100:100:20]).each(MockPar.skewed_parabola)
        data1 = loop.run(name='MockLoop_hdf5_test',
                         formatter=self.formatter)
        data2 = DataSet(location=data1.location, formatter=self.formatter)
        data2.read()
        for key in data2.arrays.keys():
            self.checkArraysEqual(data2.arrays[key], data1.arrays[key])

        metadata_equal, err_msg = compare_dictionaries(
            data1.metadata, data2.metadata,
            'original_metadata', 'loaded_metadata')
        self.assertTrue(metadata_equal, msg='\n'+err_msg)
        self.formatter.close_file(data1)
        self.formatter.close_file(data2)
        MockPar.close()

    def test_partial_dataset(self):
        data = qcodes.data.data_set.new_data(formatter=self.formatter)
        data_array = qcodes.data.data_array.DataArray(array_id='test_partial_dataset', shape=(10,))
        data_array.init_data()
        data_array.ndarray[0] = 1
        data.add_array(data_array)
        data.write()
        data.read()

    def test_loop_writing_2D(self):
        # pass
        station = Station()
        MockPar = MockParabola(name='Loop_writing_test_2D')
        station.add_component(MockPar)
        loop = Loop(MockPar.x[-100:100:20]).loop(
            MockPar.y[-50:50:10]).each(MockPar.skewed_parabola)
        data1 = loop.run(name='MockLoop_hdf5_test',
                         formatter=self.formatter)
        data2 = DataSet(location=data1.location, formatter=self.formatter)
        data2.read()
        for key in data2.arrays.keys():
            self.checkArraysEqual(data2.arrays[key], data1.arrays[key])

        metadata_equal, err_msg = compare_dictionaries(
            data1.metadata, data2.metadata,
            'original_metadata', 'loaded_metadata')
        self.assertTrue(metadata_equal, msg='\n'+err_msg)
        self.formatter.close_file(data1)
        self.formatter.close_file(data2)
        MockPar.close()

    def test_closed_file(self):
        data = DataSet1D(location=self.loc_provider, name='test_closed')
        # closing before file is written should not raise error
        self.formatter.close_file(data)
        self.formatter.write(data)
        # Used because the formatter has no nice find file method
        self.formatter.close_file(data)
        # Closing file twice should not raise an error
        self.formatter.close_file(data)

    def test_reading_into_existing_data_array(self):
        data = DataSet1D(location=self.loc_provider,
                         name='test_read_existing')
        # closing before file is written should not raise error
        self.formatter.write(data)

        data2 = DataSet(location=data.location,
                        formatter=self.formatter)
        d_array = DataArray(name='dummy',
                            array_id='x_set',  # existing array id in data
                            label='bla', unit='a.u.', is_setpoint=False,
                            set_arrays=(), preset_data=np.zeros(5))
        data2.add_array(d_array)
        # test if d_array refers to same as array x_set in dataset
        self.assertTrue(d_array is data2.arrays['x_set'])
        data2.read()
        # test if reading did not overwrite dataarray
        self.assertTrue(d_array is data2.arrays['x_set'])
        # Testing if data was correctly updated into dataset
        self.checkArraysEqual(data2.arrays['x_set'], data.arrays['x_set'])
        self.checkArraysEqual(data2.arrays['y'], data.arrays['y'])
        self.formatter.close_file(data)
        self.formatter.close_file(data2)

    def test_dataset_closing(self):
        data = DataSet1D(location=self.loc_provider, name='test_closing')
        self.formatter.write(data, flush=False)
        fp = data._h5_base_group.filename
        self.formatter.close_file(data)
        fp3 = fp[:-5]+'_3.hdf5'
        copy(fp, fp3)
        # Should not raise an error because the file was properly closed
        F3 = h5py.File(fp3, mode='a')

    def test_dataset_flush_after_write(self):
        data = DataSet1D(name='test_flush', location=self.loc_provider)
        self.formatter.write(data, flush=True)
        fp = data._h5_base_group.filename
        fp2 = fp[:-5]+'_2.hdf5'
        # the file cannot be copied unless the ref is deleted first
        del data._h5_base_group
        copy(fp, fp2)
        # Opening this copy should not raise an error
        F2 = h5py.File(fp2, mode='a')

    def test_dataset_finalize_closes_file(self):
        data = DataSet1D(name='test_finalize', location=self.loc_provider)
        # closing before file is written should not raise error
        self.formatter.write(data, flush=False)
        fp = data._h5_base_group.filename

        # Attaching the formatter like this should not be neccesary
        data.formatter = self.formatter
        data.finalize()
        # the file cannot be copied unless the ref is deleted first
        del data._h5_base_group
        fp3 = fp[:-5]+'_4.hdf5'
        copy(fp, fp3)
        # Should now not raise an error because the file was properly closed
        F3 = h5py.File(fp3, mode='a')

    def test_double_closing_gives_warning(self):
        data = DataSet1D(name='test_double_close',
                         location=self.loc_provider)
        # closing before file is written should not raise error
        self.formatter.write(data, flush=False)
        self.formatter.close_file(data)
        with self.assertLogs():
            # Test that this raises a logging message
            self.formatter.close_file(data)

    def test_dataset_with_missing_attrs(self):
        data1 = new_data(formatter=self.formatter, location=self.loc_provider,
                         name='test_missing_attr')
        arr = DataArray(array_id='arr', preset_data=np.linspace(0, 10, 21))
        data1.add_array(arr)
        data1.write()
        # data2 = DataSet(location=data1.location, formatter=self.formatter)
        # data2.read()
        data2 = load_data(location=data1.location,
                          formatter=self.formatter)
        # cannot use the check arrays equal as I expect the attributes
        # to not be equal
        np.testing.assert_array_equal(data2.arrays['arr'], data1.arrays['arr'])

    def test_read_writing_dicts_withlists_to_hdf5(self):
        some_dict = {}
        some_dict['list_of_ints'] = list(np.arange(5))
        some_dict['list_of_floats'] = list(np.arange(5.1))
        some_dict['list_of_mixed_type'] = list([1, '1'])
        fp = self.loc_provider(
            io=DataSet.default_io,
            record={'name': 'test_dict_writing'})+'.hdf5'
        F = h5py.File(fp, mode='a')

        self.formatter.write_dict_to_hdf5(some_dict, F)
        new_dict = {}
        self.formatter.read_dict_from_hdf5(new_dict, F)
        dicts_equal, err_msg = compare_dictionaries(
            some_dict, new_dict,
            'written_dict', 'loaded_dict')
        self.assertTrue(dicts_equal, msg='\n'+err_msg)

    def test_str_to_bool(self):
        self.assertEqual(str_to_bool('True'), True)
        self.assertEqual(str_to_bool('False'), False)
        with self.assertRaises(ValueError):
            str_to_bool('flse')

    def test_writing_unsupported_types_to_hdf5(self):
        """
        Tests writing of
            - unsuported list type attr
            - nested dataset
        """
        some_dict = {}
        some_dict['list_of_ints'] = list(np.arange(5))
        some_dict['list_of_floats'] = list(np.arange(5.1))
        some_dict['weird_dict'] = {'a': 5}
        data1 = new_data(formatter=self.formatter, location=self.loc_provider,
                         name='test_missing_attr')
        some_dict['nested_dataset'] = data1

        some_dict['list_of_dataset'] = [data1, data1]

        fp = self.loc_provider(
            io=DataSet.default_io,
            record={'name': 'test_dict_writing'})+'.hdf5'
        F = h5py.File(fp, mode='a')
        self.formatter.write_dict_to_hdf5(some_dict, F)
        new_dict = {}
        self.formatter.read_dict_from_hdf5(new_dict, F)
        # objects are not identical but the string representation should be
        self.assertEqual(str(some_dict['nested_dataset']),
                         new_dict['nested_dataset'])
        self.assertEqual(str(some_dict['list_of_dataset']),
                         new_dict['list_of_dataset'])

        F['weird_dict'].attrs['list_type'] = 'unsuported_list_type'
        with self.assertRaises(NotImplementedError):
            self.formatter.read_dict_from_hdf5(new_dict, F)

    def test_writing_metadata(self):
        # test for issue reported in 442
        data = DataSet2D(location=self.loc_provider, name='MetaDataTest')
        data.metadata = {'a': ['hi', 'there']}
        self.formatter.write(data, write_metadata=True)



from .KtM960xDefs import *

import ctypes
from functools import partial
from typing import (Dict, Optional, Any, Tuple)

from qcodes.instrument.base import Instrument
from qcodes.instrument.parameter import ParamRawDataType, MultiParameter
from qcodes.utils import validators as vals
from qcodes.utils.helpers import create_on_off_val_mapping


class Measure(MultiParameter):
    def __init__(self, name: str, instrument: "KtM960x") -> None:
        super().__init__(name=name,
                         names=("voltage", "current", "resistance", "status",
                                "timestamp", "source"),
                         shapes=((), (), (), (), (), ()),
                         units=("V", "A", "Ohm", "", "", ""),
                         instrument=instrument,
                         labels="Measurement Data",
                         docstring="param that returns measurement values")
        self.instrument: "KtM960x"

    def get_raw(self) -> Tuple[ParamRawDataType, ...]:
        return self.instrument._measure()


class KtM960x(Instrument):
    """
    Provide a wrapper for the Keysight KtM960x DAC. This driver provides an
    interface into the IVI-C driver provided by Keysight. The .dll is installed
    by default into C:\\Program Files\\IVI Foundation\\IVI\\Bin\\KtM960x_64.dll
    but a different path can be supplied to the constructor
    """

    _default_buf_size = 256

    def __init__(self,
                 name: str,
                 address: str,
                 options: str = "",
                 dll_path: str = r"C:\Program Files\IVI "
                                 r"Foundation\IVI\Bin\KtM960x_64.dll",
                 **kwargs: Any) -> None:
        super().__init__(name, **kwargs)

        self._address = bytes(address, "ascii")
        self._options = bytes(options, "ascii")
        self._session = ctypes.c_int(0)
        self._dll_loc = dll_path
        self._dll = ctypes.cdll.LoadLibrary(self._dll_loc)

        self.add_parameter('output',
                           label="Source Output Enable",
                           get_cmd=partial(self._get_vi_bool,
                                           KTM960X_ATTR_OUTPUT_ENABLED),
                           set_cmd=partial(self._set_vi_bool,
                                           KTM960X_ATTR_OUTPUT_ENABLED),
                           val_mapping=create_on_off_val_mapping(on_val=True,
                                                                 off_val=False)
                           )

        self.add_parameter('voltage_level',
                           label="Source Voltage Level",
                           unit="Volt",
                           get_cmd=partial(self._get_vi_real64,
                                           KTM960X_ATTR_OUTPUT_VOLTAGE_LEVEL),
                           set_cmd=partial(self._set_vi_real64,
                                           KTM960X_ATTR_OUTPUT_VOLTAGE_LEVEL),
                           vals=vals.Numbers(-210, 210))

        self.add_parameter("current_range",
                           label="Output Current Range",
                           unit="Amp",
                           vals=vals.Numbers(1e-9, 300e-3),
                           get_cmd=partial(self._get_vi_real64,
                                           KTM960X_ATTR_OUTPUT_CURRENT_RANGE),
                           set_cmd=partial(self._set_vi_real64,
                                           KTM960X_ATTR_OUTPUT_CURRENT_RANGE)
                           )

        self.add_parameter("measure_current_range",
                           label="Current Measurement Range",
                           unit="Amp",
                           get_cmd=partial(
                               self._get_vi_real64,
                               KTM960X_ATTR_MEASUREMENT_CURRENT_RANGE),
                           set_cmd=partial(
                               self._set_vi_real64,
                               KTM960X_ATTR_MEASUREMENT_CURRENT_RANGE),
                           vals=vals.Numbers(1e-9, 300e-3),
                           )

        self.add_parameter("measure_current_time",
                           label="Current Measurement Integration Time",
                           unit="Seconds",
                           get_cmd=partial(
                               self._get_vi_real64,
                               KTM960X_ATTR_MEASUREMENT_CURRENT_APERTURE),
                           set_cmd=partial(
                               self._set_vi_real64,
                               KTM960X_ATTR_MEASUREMENT_CURRENT_APERTURE),
                           vals=vals.Numbers(800e-9, 2)
                           )

        self.add_parameter("measure_data",
                           parameter_class=Measure)

        self._get_driver_desc = partial(
            self._get_vi_string, KTM960X_ATTR_SPECIFIC_DRIVER_DESCRIPTION)
        self._get_driver_prefix = partial(
            self._get_vi_string, KTM960X_ATTR_SPECIFIC_DRIVER_PREFIX)
        self._get_driver_revision = partial(
            self._get_vi_string, KTM960X_ATTR_SPECIFIC_DRIVER_REVISION)
        self._get_firmware_revision = partial(
            self._get_vi_string, KTM960X_ATTR_INSTRUMENT_FIRMWARE_REVISION)
        self._get_model = partial(
            self._get_vi_string, KTM960X_ATTR_INSTRUMENT_MODEL)
        self._get_serial_number = partial(
            self._get_vi_string, KTM960X_ATTR_MODULE_SERIAL_NUMBER)
        self._get_manufacturer = partial(
            self._get_vi_string, KTM960X_ATTR_INSTRUMENT_MANUFACTURER)

        self._connect()

        self.connect_message()

    def _connect(self) -> None:
        status = self._dll.KtM960x_InitWithOptions(self._address,
                                                   1,
                                                   1,
                                                   self._options,
                                                   ctypes.byref(self._session))
        if status:
            raise SystemError(f"connection to device failed! error: {status}")

    def get_idn(self) -> Dict[str, Optional[str]]:
        """generates the ``*IDN`` dictionary for qcodes"""

        id_dict: Dict[str, Optional[str]] = {
            'firmware': self._get_firmware_revision(),
            'model': self._get_model(),
            'serial': self._get_serial_number(),
            'vendor': self._get_manufacturer(),
            'driver desc': self._get_driver_desc(),
            'driver prefix': self._get_driver_prefix(),
            'driver revision': self._get_driver_revision()
        }
        return id_dict

    def _measure(self) -> Tuple[ParamRawDataType, ...]:

        # Setup the output
        self._set_vi_int(KTM960X_ATTR_OUTPUT_PRIORITY_MODE,
                         KTM960X_VAL_PRIORITY_MODE_VOLTAGE)
        self._set_vi_int(KTM960X_ATTR_OUTPUT_OPERATION_MODE,
                         KTM960X_VAL_OUTPUT_OPERATION_MODE_STANDARD)
        self._set_vi_int(KTM960X_ATTR_MEASUREMENT_ACQUISITION_MODE,
                         KTM960X_VAL_ACQUISITION_MODE_NORMAL)

        ch_num_buf = (ctypes.c_int32 * 1)()
        val_buf = (ctypes.c_double * 1024)()
        actual_size = ctypes.c_int32(0)
        ch_num_buf[0] = 1
        status = self._dll.KtM960x_MeasurementMeasure(
            self._session,
            KTM960X_VAL_MEASUREMENT_TYPE_ALL,
            1,
            ch_num_buf,
            1024,
            val_buf,
            ctypes.byref(actual_size)
        )

        if status:
            raise ValueError(f"Driver error: {status}")
        # This might be a bit slow?
        # Returned as [voltage, current, resistance, status,
        #                                               timestamp, and source]
        v = list(val_buf)[0:actual_size.value]

        # 'voltage': v[0], 'current': v[1], 'resistance': v[2], 'status': v[3],
        # 'timestamp': v[4], 'source': v[5]
        return v[0], v[1], v[2], v[3], v[4], v[5]

    # Query the driver for errors
    def get_errors(self) -> Dict[int, str]:
        error_code = ctypes.c_int(-1)
        error_message = ctypes.create_string_buffer(256)
        error_dict = dict()
        while error_code.value != 0:
            status = self._dll.KtM960x_error_query(
                self._session, ctypes.byref(error_code), error_message)
            assert(status == 0)
            error_dict[error_code.value] = error_message.value.decode('utf-8')

        return error_dict

    # Generic functions for reading/writing different attributes
    def _get_vi_string(self, attr: int) -> str:
        s = ctypes.create_string_buffer(self._default_buf_size)
        status = self._dll.KtM960x_GetAttributeViString(self._session,
                                                        b"",
                                                        attr,
                                                        self._default_buf_size,
                                                        s)
        if status:
            raise ValueError(f"Driver error: {status}")
        return s.value.decode('utf-8')

    def _get_vi_bool(self, attr: int) -> bool:
        s = ctypes.c_uint16(0)
        status = self._dll.KtM960x_GetAttributeViBoolean(self._session, b"",
                                                         attr, ctypes.byref(s))
        if status:
            raise ValueError(f"Driver error: {status}")
        return bool(s)

    def _set_vi_bool(self, attr: int, value: bool) -> None:
        v = ctypes.c_uint16(1) if value else ctypes.c_uint16(0)
        status = self._dll.KtM960x_SetAttributeViBoolean(self._session, b"",
                                                         attr, v)
        if status:
            raise ValueError(f"Driver error: {status}")

    def _get_vi_real64(self, attr: int) -> float:
        s = ctypes.c_double(0)
        status = self._dll.KtM960x_GetAttributeViReal64(self._session, b"",
                                                        attr, ctypes.byref(s))

        if status:
            raise ValueError(f"Driver error: {status}")
        return float(s.value)

    def _set_vi_real64(self, attr: int, value: float) -> None:
        v = ctypes.c_double(value)
        status = self._dll.KtM960x_SetAttributeViReal64(self._session, b"",
                                                        attr, v)
        if status:
            raise ValueError(f"Driver error: {status}")

    def _set_vi_int(self, attr: int, value: int) -> None:
        v = ctypes.c_int32(value)
        status = self._dll.KtM960x_SetAttributeViInt32(self._session, b"",
                                                       attr, v)
        if status:
            raise ValueError(f"Driver error: {status}")

    def _get_vi_int(self, attr: int) -> int:
        v = ctypes.c_int32(0)
        status = self._dll.KtM960x_GetAttributeViInt32(self._session, b"",
                                                       attr, ctypes.byref(v))
        if status:
            raise ValueError(f"Driver error: {status}")
        return int(v.value)

    def close(self) -> None:
        self._dll.KtM960x_close(self._session)
        super().close()



from distutils.version import LooseVersion
from typing import Any, Sequence, Tuple, Union, cast

from pyvisa.errors import VisaIOError

from qcodes import InstrumentChannel, VisaInstrument
from qcodes.instrument.group_parameter import Group, GroupParameter
from qcodes.instrument.parameter import (
    ManualParameter,
    MultiParameter,
    ParamRawDataType,
)
from qcodes.utils.helpers import create_on_off_val_mapping
from qcodes.utils.validators import Bool, Enum, Ints, Numbers


class MeasurementPair(MultiParameter):
    """
    Data class for E4980A measurement, which will always return two items
    at once.

    The two items are for two different parameters, depending on the measurement
    function. Hence, the names of the two attributes are created from the
    "names" tuple of the measurement functions.

    Examples:
        To create a measurement data with capacitance=1.2, and
        dissipation_factor=3.4.

        >>> data = MeasurementPair(name="CPD",
                                    names=("capacitance", "dissipation_factor"),
                                    units=("F", ""))
        >>> data.set((1.2, 3.4))
        >>> data.get()
        (1.2, 3.4)
    """
    value: Tuple[float, float] = (0., 0.)

    def __init__(self,
                 name: str,
                 names: Sequence[str],
                 units: Sequence[str],
                 **kwargs: Any):
        super().__init__(name=name,
                         names=names,
                         shapes=((), ()),
                         units=units,
                         setpoints=((), ()),
                         **kwargs)
        self.__dict__.update(
            {names[0]: 0,
             names[1]: 0}
        )

    def set_raw(self, value: Tuple[float, float]) -> None:
        self.value = value
        setattr(self, self.names[0], value[0])
        setattr(self, self.names[1], value[1])

    def get_raw(self) -> Tuple[ParamRawDataType, ...]:
        return self.value


class E4980AMeasurements:
    """
    All the measurement function for E4980A LCR meter. See user's guide P353
    https://literature.cdn.keysight.com/litweb/pdf/E4980-90230.pdf?id=789356
    """
    CPD = MeasurementPair(
        "CPD", ("capacitance", "dissipation_factor"), ("F", "")
    )
    CPQ = MeasurementPair(
        "CPQ", ("capacitance", "quality_factor"), ("F", "")
    )
    CPG = MeasurementPair(
        "CPG", ("capacitance", "conductance"), ("F", "S")
    )
    CPRP = MeasurementPair(
        "CPRP", ("capacitance", "resistance"), ("F", "Ohm")
    )
    CSD = MeasurementPair(
        "CSD", ("capacitance", "dissipation_factor"), ("F", "")
    )
    CSQ = MeasurementPair(
        'CSQ', ("capacitance", "quality_factor"), ("F", "")
    )
    CSRS = MeasurementPair(
        "CSRS", ("capacitance", "resistance"), ("F", "Ohm")
    )
    LPD = MeasurementPair(
        "LPD", ("inductance", "dissipation_factor"), ("H", "")
    )
    LPQ = MeasurementPair(
        "LPQ", ("inductance", "quality_factor"), ("H", "")
    )
    LPG = MeasurementPair(
        "LPG", ("inductance", "conductance"), ("H", "S")
    )
    LPRP = MeasurementPair(
        "LPRP", ("inductance", "resistance"), ("H", "Ohm")
    )
    LSD = MeasurementPair(
        "LSD", ("inductance", "dissipation_factor"), ("H", "")
    )
    LSQ = MeasurementPair(
        "LSQ", ("inductance", "quality_factor"), ("H", "")
    )
    LSRS = MeasurementPair(
        "LSRS", ("inductance", "resistance"), ("H", "Ohm")
    )
    LSRD = MeasurementPair(
        "LSRD", ("inductance", "resistance"), ("H", "Ohm")
    )
    RX = MeasurementPair(
        "RX", ("resistance", "reactance"), ("Ohm", "Ohm")
    )
    ZTD = MeasurementPair(
        "ZTD", ("impedance", "theta"), ("Ohm", "Degree")
    )
    ZTR = MeasurementPair(
        "ZTR", ("impedance", "theta"), ("Ohm", "Radiant")
    )
    GB = MeasurementPair(
        "GB", ("conductance", "susceptance"), ("S", "S")
    )
    YTD = MeasurementPair(
        "YTD", ("admittance", "theta"), ("Y", "Degree")
    )
    YTR = MeasurementPair(
        "YTR", ("admittance", "theta"), ("Y", "Radiant")
    )
    VDID = MeasurementPair(
        "VDID", ("voltage", "current"), ("V", "A")
    )


class Correction4980A(InstrumentChannel):
    """
    Module for correction settings.
    """
    def __init__(
            self,
            parent: VisaInstrument,
            name: str,
    ) -> None:
        super().__init__(parent, name)

        self.add_parameter(
            "open",
            set_cmd=":CORRection:OPEN",
            docstring="Executes OPEN correction based on all frequency points."
        )

        self.add_parameter(
            "open_state",
            get_cmd=":CORRection:OPEN:STATe?",
            set_cmd=":CORRection:OPEN:STATe {}",
            val_mapping=create_on_off_val_mapping(on_val="1", off_val="0"),
            docstring="Enables or disable OPEN correction"
        )

        self.add_parameter(
            "short",
            set_cmd=":CORRection:SHORt",
            docstring="Executes SHORT correction based on all frequency points."
        )

        self.add_parameter(
            "short_state",
            get_cmd=":CORRection:SHORt:STATe?",
            set_cmd=":CORRection:SHORt:STATe {}",
            val_mapping=create_on_off_val_mapping(on_val="1", off_val="0"),
            docstring="Enables or disable SHORT correction."
        )


class KeysightE4980A(VisaInstrument):
    """
    QCodes driver for E4980A Precision LCR Meter
    """
    def __init__(self,
                 name: str,
                 address: str,
                 terminator: str = '\n',
                 **kwargs: Any):
        """
        Create an instance of the instrument.

        Args:
            name: Name of the instrument instance
            address: Visa-resolvable instrument address.
        """
        super().__init__(name, address, terminator=terminator, **kwargs)

        idn = self.IDN.get()

        self.has_firmware_a_02_10_or_above = (
                LooseVersion(idn["firmware"]) >= LooseVersion("A.02.10")
        )

        self.has_option_001 = '001' in self._options()
        self._dc_bias_v_level_range: Union[Numbers, Enum]
        if self.has_option_001:
            self._v_level_range = Numbers(0, 20)
            self._i_level_range = Numbers(0, 0.1)
            self._imp_range = Enum(0.1, 1, 10, 100, 300, 1000, 3000, 10000,
                                   30000, 100000)
            self._dc_bias_v_level_range = Numbers(-40, 40)
        else:
            self._v_level_range = Numbers(0, 2)
            self._i_level_range = Numbers(0, 0.02)
            self._imp_range = Enum(1, 10, 100, 300, 1000, 3000, 10000, 30000,
                                   100000)
            self._dc_bias_v_level_range = Enum(0, 1.5, 2)

        self._measurement_pair = MeasurementPair(
            "CPD",
            ("capacitance", "dissipation_factor"),
            ("F", "")
        )

        self.add_parameter(
            "frequency",
            get_cmd=":FREQuency?",
            set_cmd=":FREQuency {}",
            get_parser=float,
            unit="Hz",
            vals=Numbers(20, 2E6),
            docstring="Gets and sets the frequency for normal measurement."
        )

        self.add_parameter(
            "current_level",
            get_cmd=self._get_current_level,
            set_cmd=self._set_current_level,
            unit="A",
            vals=self._i_level_range,
            docstring="Gets and sets the current level for measurement signal."
        )

        self.add_parameter(
            "voltage_level",
            get_cmd=self._get_voltage_level,
            set_cmd=self._set_voltage_level,
            unit="V",
            vals=self._v_level_range,
            docstring="Gets and sets the AC bias voltage level for measurement "
                      "signal."
        )

        self.add_parameter(
            "measurement_function",
            get_cmd=":FUNCtion:IMPedance?",
            set_cmd=self._set_measurement
        )

        self.add_parameter(
            "range",
            get_cmd=":FUNCtion:IMPedance:RANGe?",
            set_cmd=self._set_range,
            unit='Ohm',
            vals=self._imp_range,
            docstring="Selects the impedance measurement range, also turns "
                      "the auto range function OFF."
        )

        self.add_parameter(
            "imp_autorange_enabled",
            get_cmd=":FUNCtion:IMPedance:RANGe:AUTO?",
            set_cmd=":FUNCtion:IMPedance:RANGe:AUTO {}",
            val_mapping=create_on_off_val_mapping(on_val="1",
                                                  off_val="0"),
            docstring="Enables the auto-range for impedance measurement."
        )

        self.add_parameter(
            "dc_bias_enabled",
            get_cmd=":BIAS:STATe?",
            set_cmd=":BIAS:STATe {}",
            vals=Bool(),
            val_mapping=create_on_off_val_mapping(on_val="1",
                                                  off_val="0"),
            docstring="Enables DC bias. DC bias is automatically turned "
                      "off after recalling the state from memory."
        )

        self.add_parameter(
            "dc_bias_voltage_level",
            get_cmd=":BIAS:VOLTage:LEVel?",
            set_cmd=":BIAS:VOLTage:LEVel {}",
            get_parser=float,
            unit="V",
            vals=self._dc_bias_v_level_range,
            docstring="Sets the DC bias voltage. Setting does not "
                      "implicitly turn the DC bias ON."
        )

        self.add_parameter(
            "meas_time_mode",
            val_mapping={"short": "SHOR", "medium": "MED", "long": "LONG"},
            parameter_class=GroupParameter
        )

        self.add_parameter(
            "averaging_rate",
            vals=Ints(1, 256),
            parameter_class=GroupParameter,
            get_parser=int,
            docstring="Averaging rate for the measurement."
        )

        self._aperture_group = Group(
            [self.meas_time_mode,
             self.averaging_rate],
            set_cmd=":APERture {meas_time_mode},{averaging_rate}",
            get_cmd=":APERture?"
        )

        if self.has_firmware_a_02_10_or_above:
            self.add_parameter(
                "dc_bias_autorange_enabled",
                get_cmd=":BIAS:RANGe:AUTO?",
                set_cmd=":BIAS:RANGe:AUTO {}",
                vals=Bool(),
                val_mapping=create_on_off_val_mapping(on_val="1",
                                                      off_val="0"),
                docstring="Enables DC Bias range AUTO setting. When DC bias "
                          "range is fixed (not AUTO), '#' is displayed in "
                          "the BIAS field of the display."
            )

        self.add_parameter(
            "signal_mode",
            initial_value=None,
            vals=Enum("Voltage", "Current", None),
            parameter_class=ManualParameter,
            docstring="This parameter tracks the signal mode which is being "
                      "set."
        )

        self.add_submodule(
            "_correction",
            Correction4980A(self, "correction")
        )
        self._set_signal_mode_on_driver_initialization()
        self.connect_message()

    @property
    def correction(self) -> Correction4980A:
        submodule = self.submodules['_correction']
        return cast(Correction4980A, submodule)

    @property
    def measure_impedance(self) -> MeasurementPair:
        return self._get_complex_impedance()

    @property
    def measurement(self) -> MeasurementPair:
        return self._measurement()

    def _set_range(self, val: str) -> None:
        self.write(f":FUNCtion:IMPedance:RANGe {val}")
        self.imp_autorange_enabled.get()

    def _get_complex_impedance(self) -> MeasurementPair:
        """
        Returns the impedance in the format of (R, X), where R is the
        resistance, and X is the reactance.
        """
        measurement = self.ask(":FETCH:IMPedance:CORRected?")
        r, x = (float(n) for n in measurement.split(","))
        measurement_pair = MeasurementPair(
            name="RX",
            names=("resistance", "reactance"),
            units=("Ohm", "Ohm")
        )
        measurement_pair.set((r, x))
        return measurement_pair

    def _measurement(self) -> MeasurementPair:
        """
        Returns a measurement result with the selected measurement function.
        """
        measurement = self.ask(":FETCH:IMPedance:FORMatted?")
        val1, val2, _ = (float(n) for n in measurement.split(","))
        measurement_pair = MeasurementPair(
            name=self._measurement_pair.name,
            names=self._measurement_pair.names,
            units=self._measurement_pair.units
        )
        measurement_pair.set((val1, val2))
        return measurement_pair

    def _set_measurement(self,
                         measurement_pair: MeasurementPair) -> None:
        """
        Selects the measurement function.
        """
        self._measurement_pair = measurement_pair
        self.write(f":FUNCtion:IMPedance {measurement_pair.name}")

    def _get_voltage_level(self) -> float:
        """
        Gets voltage level if signal is set with voltage level parameter
        otherwise raises an error.
        """
        if self.signal_mode() == "Current":
            raise RuntimeError("Cannot get voltage level as signal is set "
                               "with current level parameter.")

        v_level = self.ask(":VOLTage:LEVel?")

        return float(v_level)

    def _set_voltage_level(self, val: str) -> None:
        """
        Sets voltage level
        """
        self.signal_mode("Voltage")
        self.voltage_level.snapshot_exclude = False
        self.current_level.snapshot_exclude = True

        self.write(f":VOLTage:LEVel {val}")

    def _set_current_level(self, val: str) -> None:
        """
        Sets current level
        """
        self.signal_mode("Current")
        self.voltage_level.snapshot_exclude = True
        self.current_level.snapshot_exclude = False

        self.write(f":CURRent:LEVel {val}")

    def _get_current_level(self) -> float:
        """
        Gets current level if signal is set with current level parameter
        otherwise raises an error.
        """
        if self.signal_mode() == "Voltage":
            raise RuntimeError("Cannot get current level as signal is set "
                               "with voltage level parameter.")

        i_level = self.ask(":CURRent:LEVel?")

        return float(i_level)

    def _is_signal_mode_voltage_on_driver_initialization(self) -> bool:
        """
        Checks if signal is set with voltage_level param at instrument driver
        initialization
        """
        assert self.signal_mode() is None
        try:
            self.voltage_level()
            return True
        except VisaIOError:
            return False

    def _set_signal_mode_on_driver_initialization(self) -> None:
        """
        Sets signal mode on driver initialization
        """
        if self._is_signal_mode_voltage_on_driver_initialization():
            self.signal_mode("Voltage")
            self.voltage_level.snapshot_exclude = False
            self.current_level.snapshot_exclude = True
        else:
            self.signal_mode("Current")
            self.voltage_level.snapshot_exclude = True
            self.current_level.snapshot_exclude = False

    def _options(self) -> Tuple[str, ...]:
        """
        Returns installed options numbers. Combinations of different installed
        options are possible. Two of the possible options are Power/DC Bias
        Enhance (option 001) and Bias Current Interface (option 002).
        """
        options_raw = self.ask('*OPT?')
        return tuple(options_raw.split(','))

    def system_errors(self) -> str:
        """
        Returns the oldest unread error message from the event log and removes
        it from the log.
        """
        return self.ask(":SYSTem:ERRor?")

    def clear_status(self) -> None:
        """
        Clears the following:
            Error Queue
            Status Byte Register
            Standard Event Status Register
            Operation Status Event Register
            Questionable Status Event Register (No Query)
        """
        self.write('*CLS')

    def reset(self) -> None:
        """
        Resets the instrument settings.
        """
        self.write('*RST')



from typing import Any, Optional, Dict

from qcodes import Parameter, DelegateParameter
from qcodes.instrument.base import Instrument
from qcodes.utils.validators import Enum, Numbers


class SP983C(Instrument):
    """
    A virtual driver for the Basel SP 983 and SP 983C current to voltage
    converter.

    This driver supports both the SP 983 and SP 983C models. These differ only
    in their handling of input offset voltage. It is the responsibility of the
    user to capture the input offset, (from the voltage supply) and compensate
    that as needed for SP 983. For SP 983C model, 'input_offset_voltage'
    argument can be used to set up offset (This doesn't work for SP 983c01
    model).

    Note that, as this is a purely virtual driver, there is no support
    for the the remote control interface (SP 983a). It is the responsibility of
    the user to ensure that values set here are in accordance with the values
    set on the instrument.

    Args:
        name
        input_offset_voltage: (Optional) A source input offset voltage
            parameter. The range for input is -10 to 10 Volts and it is
            user's responsibility to ensure this. This source parameter is
            used to set offset voltage parameter of the preamp and the
            source parameter should represent a voltage source that is
            connected to the "Offset Input Volgate" connector of the SP983C.
    """

    def __init__(self, name: str,
                 input_offset_voltage: Optional[Parameter] = None,
                 **kwargs: Any):
        super().__init__(name, **kwargs)

        self.add_parameter('gain',
                           initial_value=1e8,
                           label='Gain',
                           unit='V/A',
                           get_cmd=None, set_cmd=None,
                           vals=Enum(1e09, 1e08, 1e07, 1e06, 1e05))

        self.add_parameter('fcut',
                           initial_value=1e3,
                           label='Cutoff frequency',
                           unit='Hz',
                           get_cmd=None, set_cmd=None,
                           vals=Enum(30., 100., 300., 1e3, 3e3, 10e3, 30e3,
                                     100e3, 1e6))

        self.add_parameter('offset_voltage',
                           label="Offset Voltage",
                           unit='V',
                           vals=Numbers(-0.1, 0.1),
                           scale=100,
                           source=input_offset_voltage,
                           parameter_class=DelegateParameter)

    def get_idn(self) -> Dict[str, Optional[str]]:
        vendor = 'Physics Basel'
        model = 'SP 983(c)'
        serial = None
        firmware = None
        return {'vendor': vendor, 'model': model,
                'serial': serial, 'firmware': firmware}



"""
This module provides a class that encapsulates Alazar ATS API,
:class:`AlazarATSAPI`. The class is used to expose Alazar API functions
of its C library in a python-friendly way.
"""


from typing import Dict, Tuple, Union, Any
import ctypes
from ctypes import POINTER

from .dll_wrapper import WrappedDll, Signature
from .constants import BOARD_NAMES, REGISTER_ACCESS_PASSWORD, ReturnCode

# `_BaseParameter` is needed because users may pass instrument parameters
# that originate from `Instrument.parameters` dictionary which is typed
# with `_BaseParameter`, not `Parameter`.
from qcodes.instrument.parameter import _BaseParameter as Parameter

# Define aliases for ctypes that match Alazar's notation.
U8 = ctypes.c_uint8
U16 = ctypes.c_uint16
U32 = ctypes.c_uint32
C_LONG = ctypes.c_long
HANDLE = ctypes.c_void_p

POINTER_c_uint8 = Any
POINTER_c_uint16 = Any
POINTER_c_uint32 = Any
POINTER_c_long = Any


class AlazarATSAPI(WrappedDll):
    """
    A thread-safe wrapper for the ATS API library.

    The exposed ATS API functions have snake_case names, may accept python
    types and :class:`qcodes.instrument.parameter.Parameter` s as input.
    The API calls are thread-safe, and are executed in a separate thread.
    Moreover, behind the scenes it is ensured that only single instance of
    this class exists per unique ATS API DLL file.

    Some of the ATS API functions are also exposed with more convenient
    signatures. These usually have the same name but with an additional
    underscore at the end.
    """

    ## CONSTANTS ##

    BOARD_NAMES = BOARD_NAMES

    ## ACTUAL DLL API FUNCTIONS ##

    signatures: Dict[str, Signature] = {}

    def set_trigger_time_out(self,
                             handle: int,
                             timeout_ticks: Union[int, Parameter]
                            ) -> ReturnCode:
        return self._sync_dll_call(
            'AlazarSetTriggerTimeOut', handle, timeout_ticks)

    signatures.update({"AlazarSetTriggerTimeOut": Signature(
        argument_types=[HANDLE, U32])})

    def get_board_kind(self, handle: int) -> int:
        return self._sync_dll_call('AlazarGetBoardKind', handle)

    signatures.update({'AlazarGetBoardKind': Signature(
        argument_types=[HANDLE],
        return_type=U32)})

    def get_channel_info(self,
                         handle: int,
                         memory_size_in_samples: POINTER_c_uint32,
                         bits_per_sample: POINTER_c_uint8
                         ) -> ReturnCode:
        return self._sync_dll_call(
            'AlazarGetChannelInfo',
            handle, memory_size_in_samples, bits_per_sample)

    signatures.update({'AlazarGetChannelInfo': Signature(
            argument_types=[HANDLE, POINTER(U32), POINTER(U8)])})

    def get_cpld_version(self,
                         handle: int,
                         major: POINTER_c_uint8,
                         minor: POINTER_c_uint8
                         ) -> ReturnCode:
        return self._sync_dll_call(
            'AlazarGetCPLDVersion', handle, major, minor)

    signatures.update({"AlazarGetCPLDVersion": Signature(
        argument_types=[HANDLE, POINTER(U8), POINTER(U8)])})

    def get_driver_version(self,
                           major: POINTER_c_uint8,
                           minor: POINTER_c_uint8,
                           revision: POINTER_c_uint8
                           ) -> ReturnCode:
        return self._sync_dll_call(
            'AlazarGetDriverVersion', major, minor, revision)

    signatures.update({'AlazarGetDriverVersion': Signature(
            argument_types=[POINTER(U8), POINTER(U8), POINTER(U8)])})

    def get_sdk_version(self,
                        major: POINTER_c_uint8,
                        minor: POINTER_c_uint8,
                        revision: POINTER_c_uint8
                        ) -> ReturnCode:
        return self._sync_dll_call(
            'AlazarGetSDKVersion', major, minor, revision)

    signatures.update({'AlazarGetSDKVersion': Signature(
            argument_types=[POINTER(U8), POINTER(U8), POINTER(U8)])})

    def query_capability(self,
                         handle: int,
                         capability: int,
                         reserved: int,
                         value: POINTER_c_uint32
                         ) -> ReturnCode:
        return self._sync_dll_call(
            'AlazarQueryCapability', handle, capability, reserved, value)

    signatures.update({'AlazarQueryCapability': Signature(
            argument_types=[HANDLE, U32, U32, POINTER(U32)])})

    def read_register(self,
                      handle: int,
                      offset: int,
                      output: POINTER_c_uint32,
                      password: int,
                      ) -> ReturnCode:
        return self._sync_dll_call(
            'AlazarReadRegister', handle, offset, output, password)

    signatures.update({"AlazarReadRegister": Signature(
            argument_types=[HANDLE, U32, POINTER(U32), U32])})

    def write_register(self,
                       handle: int,
                       offset: int,
                       value: int,
                       password: int
                      ) -> ReturnCode:
        return self._sync_dll_call(
            'AlazarWriteRegister', handle, offset, value, password)

    signatures.update({"AlazarWriteRegister": Signature(
            argument_types=[HANDLE, U32, U32, U32])})

    def num_of_systems(self) -> int:
        return self._sync_dll_call('AlazarNumOfSystems')

    signatures.update({'AlazarNumOfSystems': Signature(
            return_type=U32)})

    def boards_in_system_by_system_id(self, system_id: int) -> int:
        return self._sync_dll_call(
            'AlazarBoardsInSystemBySystemID', system_id)

    signatures.update({'AlazarBoardsInSystemBySystemID': Signature(
        argument_types=[U32], return_type=U32)})

    def get_board_by_system_id(self, system_id: int, board_id: int) -> int:
        return self._sync_dll_call(
            'AlazarGetBoardBySystemID', system_id, board_id)

    signatures.update({'AlazarGetBoardBySystemID': Signature(
        argument_types=[U32, U32], return_type=HANDLE)})

    def get_parameter(self, handle: int, channel: int,
                      parameter: int,
                      ret_value: POINTER_c_long) -> ReturnCode:
        return self._sync_dll_call(
            'AlazarGetParameter', handle, channel,
            parameter, ret_value)

    signatures.update({'AlazarGetParameter': Signature(
        argument_types=[HANDLE, U8, U32, POINTER(C_LONG)])})

    def set_parameter(self, handle: int, channel: int,
                      parameter: int,
                      value: int) -> ReturnCode:
        return self._sync_dll_call(
            'AlazarSetParameter', handle, channel,
            parameter, value)

    signatures.update({'AlazarSetParameter': Signature(
        argument_types=[HANDLE, U8, U32, C_LONG])})

    def set_capture_clock(self,
                          handle: int,
                          source_id: Union[int, Parameter],
                          sample_rate_id_or_value: Union[int, Parameter],
                          edge_id: Union[int, Parameter],
                          decimation: Union[int, Parameter]
                          ) -> ReturnCode:
        return self._sync_dll_call(
            'AlazarSetCaptureClock',
            handle, source_id, sample_rate_id_or_value, edge_id, decimation)

    signatures.update({"AlazarSetCaptureClock": Signature(
        argument_types=[HANDLE, U32, U32, U32, U32])})

    def input_control(self,
                      handle: int,
                      channel_id: Union[int, Parameter],
                      coupling_id: Union[int, Parameter],
                      range_id: Union[int, Parameter],
                      impedance_id: Union[int, Parameter]
                      ) -> ReturnCode:
        return self._sync_dll_call(
            'AlazarInputControl',
            handle, channel_id, coupling_id, range_id, impedance_id)

    signatures.update({"AlazarInputControl": Signature(
        argument_types=[HANDLE, U8, U32, U32, U32])})

    def set_bw_limit(self,
                     handle: int,
                     channel_id: Union[int, Parameter],
                     flag: Union[int, Parameter]
                     ) -> ReturnCode:
        return self._sync_dll_call(
            'AlazarSetBWLimit', handle, channel_id, flag)

    signatures.update({"AlazarSetBWLimit": Signature(
        argument_types=[HANDLE, U32, U32])})

    def set_trigger_operation(self,
                              handle: int,
                              trigger_operation: Union[int, Parameter],
                              trigger_engine_id_1: Union[int, Parameter],
                              source_id_1: Union[int, Parameter],
                              slope_id_1: Union[int, Parameter],
                              level_1: Union[int, Parameter],
                              trigger_engine_id_2: Union[int, Parameter],
                              source_id_2: Union[int, Parameter],
                              slope_id_2: Union[int, Parameter],
                              level_2: Union[int, Parameter],
                              ) -> ReturnCode:
        return self._sync_dll_call(
            'AlazarSetTriggerOperation',
            handle, trigger_operation,
            trigger_engine_id_1, source_id_1, slope_id_1, level_1,
            trigger_engine_id_2, source_id_2, slope_id_2, level_2)

    signatures.update({"AlazarSetTriggerOperation": Signature(
        argument_types=[
            HANDLE, U32, U32, U32, U32, U32, U32, U32, U32, U32])})

    def set_external_trigger(self,
                             handle: int,
                             coupling_id: Union[int, Parameter],
                             range_id: Union[int, Parameter]
                             ) -> ReturnCode:
        return self._sync_dll_call(
            'AlazarSetExternalTrigger', handle, coupling_id, range_id)

    signatures.update({"AlazarSetExternalTrigger": Signature(
        argument_types=[HANDLE, U32, U32])})

    def set_trigger_delay(self,
                          handle: int,
                          value: Union[int, Parameter]
                          ) -> ReturnCode:
        return self._sync_dll_call('AlazarSetTriggerDelay', handle, value)

    signatures.update({"AlazarSetTriggerDelay": Signature(
        argument_types=[HANDLE, U32])})

    def configure_aux_io(self,
                         handle: int,
                         mode_id: Union[int, Parameter],
                         mode_parameter_value: Union[int, Parameter]
                         ) -> ReturnCode:
        return self._sync_dll_call(
            'AlazarConfigureAuxIO', handle, mode_id, mode_parameter_value)

    signatures.update({"AlazarConfigureAuxIO": Signature(
        argument_types=[HANDLE, U32, U32])})

    def set_record_size(self,
                        handle: int,
                        pre_trigger_samples: Union[int, Parameter],
                        post_trigger_samples: Union[int, Parameter]
                        ) -> ReturnCode:
        return self._sync_dll_call(
            'AlazarSetRecordSize', handle, pre_trigger_samples, post_trigger_samples)

    signatures.update({"AlazarSetRecordSize": Signature(
        argument_types=[HANDLE, U32, U32])})

    def before_async_read(self,
                          handle: int,
                          channel_select: int,
                          transfer_offset: int,
                          samples_per_record: int,
                          records_per_buffer: int,
                          records_per_acquisition: int,
                          flags: int
                          ) -> ReturnCode:
        return self._sync_dll_call(
            'AlazarBeforeAsyncRead',
            handle, channel_select, transfer_offset,
            samples_per_record, records_per_buffer, records_per_acquisition,
            flags)

    signatures.update({"AlazarBeforeAsyncRead": Signature(
        argument_types=[HANDLE, U32, C_LONG, U32, U32, U32, U32])})

    def post_async_buffer(self,
                          handle: int,
                          buffer: ctypes.c_void_p,
                          buffer_length: int
                          ) -> ReturnCode:
        return self._sync_dll_call(
            'AlazarPostAsyncBuffer', handle, buffer, buffer_length)

    signatures.update({"AlazarPostAsyncBuffer": Signature(
        argument_types=[HANDLE, ctypes.c_void_p, U32])})

    def wait_async_buffer_complete(self,
                                   handle: int,
                                   buffer: ctypes.c_void_p,
                                   timeout_in_ms: int
                                   ) -> ReturnCode:
        return self._sync_dll_call(
            'AlazarWaitAsyncBufferComplete', handle, buffer, timeout_in_ms)

    signatures.update({"AlazarWaitAsyncBufferComplete": Signature(
        argument_types=[HANDLE, ctypes.c_void_p, U32])})

    def start_capture(self, handle: int) -> ReturnCode:
        return self._sync_dll_call('AlazarStartCapture', handle)

    signatures.update({"AlazarStartCapture": Signature(
        argument_types=[HANDLE])})

    def abort_async_read(self, handle: int) -> ReturnCode:
        return self._sync_dll_call('AlazarAbortAsyncRead', handle)

    signatures.update({"AlazarAbortAsyncRead": Signature(
        argument_types=[HANDLE])})

    def error_to_text(self, return_code: ReturnCode) -> str:
        return self._sync_dll_call('AlazarErrorToText', return_code)

    signatures.update({"AlazarErrorToText": Signature(
        argument_types=[U32], return_type=ctypes.c_char_p)})

    def force_trigger(self, handle: int) -> ReturnCode:
        return self._sync_dll_call('AlazarForceTrigger', handle)

    signatures.update({"AlazarForceTrigger": Signature(
        argument_types=[HANDLE])})

    def force_trigger_enable(self, handle: int) -> ReturnCode:
        return self._sync_dll_call('AlazarForceTriggerEnable', handle)

    signatures.update({"AlazarForceTriggerEnable": Signature(
        argument_types=[HANDLE])})

    def busy(self, handle: int) -> int:
        return self._sync_dll_call('AlazarBusy', handle)

    signatures.update({"AlazarBusy": Signature(
        argument_types=[HANDLE], return_type=U32)})

    def configure_record_average(self,
                                 handle: int,
                                 mode: int,
                                 samples_per_record: int,
                                 records_per_average: int,
                                 options: int
                                 ) -> ReturnCode:
        return self._sync_dll_call(
            'AlazarConfigureRecordAverage',
            handle, mode, samples_per_record, records_per_average, options)

    signatures.update({"AlazarConfigureRecordAverage": Signature(
        argument_types=[HANDLE, U32, U32, U32, U32])})

    def free_buffer_u16(self,
                        handle: int,
                        buffer: POINTER_c_uint16
                        ) -> ReturnCode:
        return self._sync_dll_call('AlazarFreeBufferU16', handle, buffer)

    signatures.update({"AlazarFreeBufferU16": Signature(
        argument_types=[HANDLE, POINTER(U16)])})

    def free_buffer_u8(self,
                       handle: int,
                       buffer: POINTER_c_uint8
                       ) -> ReturnCode:
        return self._sync_dll_call('AlazarFreeBufferU8', handle, buffer)

    signatures.update({"AlazarFreeBufferU8": Signature(
        argument_types=[HANDLE, POINTER(U8)])})

    def set_led(self, handle: int, led_on: bool) -> ReturnCode:
        return self._sync_dll_call('AlazarSetLED', handle, led_on)

    signatures.update({"AlazarSetLED": Signature(
        argument_types=[HANDLE, U32])})

    def triggered(self, handle: int) -> int:
        return self._sync_dll_call('AlazarTriggered', handle)

    signatures.update({"AlazarTriggered": Signature(
        argument_types=[HANDLE], return_type=U32)})

    ## OTHER API-RELATED METHODS ##

    def get_board_model(self, handle: int) -> str:
        return self.BOARD_NAMES[self.get_board_kind(handle)]

    def get_channel_info_(self, handle: int) -> Tuple[int, int]:
        """
        A more convenient version of :meth:`get_channel_info` method
        (``AlazarGetChannelInfo``).

        This method hides the fact that the output values in the original
        function are written to the provided pointers.

        Args:
            handle: Handle of the board of interest

        Returns:
            Tuple of bits per sample and maximum board memory in samples
        """
        bps = ctypes.c_uint8(0)  # bps bits per sample
        max_s = ctypes.c_uint32(0)  # max_s memory size in samples
        self.get_channel_info(
            handle,
            ctypes.byref(max_s),
            ctypes.byref(bps)
        )
        return max_s.value, bps.value

    def get_cpld_version_(self, handle: int) -> str:
        """
        A more convenient version of :meth:`get_cpld_version` method
        (``AlazarGetCPLDVersion``).

        This method hides the fact that the output values in the original
        function are written to the provided pointers.

        Args:
            handle: Handle of the board of interest

        Returns:
            Version string in the format "<major>.<minor>"
        """
        major = ctypes.c_uint8(0)
        minor = ctypes.c_uint8(0)
        self.get_cpld_version(
            handle,
            ctypes.byref(major),
            ctypes.byref(minor)
        )
        cpld_ver = str(major.value) + "." + str(minor.value)
        return cpld_ver

    def get_driver_version_(self) -> str:
        """
        A more convenient version of :meth:`get_driver_version` method
        (``AlazarGetDriverVersion``).

        This method hides the fact that the output values in the original
        function are written to the provided pointers.

        Returns:
            Version string in the format "<major>.<minor>.<revision>"
        """
        major = ctypes.c_uint8(0)
        minor = ctypes.c_uint8(0)
        revision = ctypes.c_uint8(0)
        self.get_driver_version(
            ctypes.byref(major),
            ctypes.byref(minor),
            ctypes.byref(revision)
        )
        driver_ver = (str(major.value) + "."
                      + str(minor.value) + "."
                      + str(revision.value))
        return driver_ver

    def get_sdk_version_(self) -> str:
        """
        A more convenient version of :meth:`get_sdk_version` method
        (``AlazarGetSDKVersion``).

        This method hides the fact that the output values in the original
        function are written to the provided pointers.

        Returns:
            Version string in the format "<major>.<minor>.<revision>"
        """
        major = ctypes.c_uint8(0)
        minor = ctypes.c_uint8(0)
        revision = ctypes.c_uint8(0)
        self.get_sdk_version(
            ctypes.byref(major),
            ctypes.byref(minor),
            ctypes.byref(revision)
        )
        sdk_ver = (str(major.value) + "."
                   + str(minor.value) + "."
                   + str(revision.value))
        return sdk_ver

    def query_capability_(self, handle: int, capability: int) -> int:
        """
        A more convenient version of :meth:`query_capability` method
        (``AlazarQueryCapability``).

        This method hides the fact that the output values in the original
        function are written to the provided pointers.

        Args:
            handle: Handle of the board of interest
            capability: An integer identifier of a capability parameter
                (:class:`.constants.Capability` enumeration encapsulates
                the available identifiers)

        Returns:
            Value of the requested capability
        """
        value = ctypes.c_uint32(0)
        reserved = 0
        self.query_capability(
            handle, capability, reserved, ctypes.byref(value))
        return value.value

    def read_register_(self, handle: int, offset: int) -> int:
        """
        Read a value from a given offset in the Alazar card's register.

        A more convenient version of :meth:`read_register` method
        (``AlazarReadRegister``).

        Args:
            handle: Handle of the board of interest
            offset: Offset into the memory to read from

        Returns:
            The value read as an integer
        """
        output = ctypes.c_uint32(0)
        self.read_register(
            handle,
            offset,
            ctypes.byref(output),
            REGISTER_ACCESS_PASSWORD
        )
        return output.value

    def write_register_(self, handle: int, offset: int, value: int) -> None:
        """
        Write a value to a given offset in the Alazar card's register.

        A more convenient version of :meth:`write_register` method
        (``AlazarWriteRegister``).

        Args:
            handle: Handle of the board of interest
            offset: The offset in memory to write to
            value: The value to write
        """
        self.write_register(
            handle, offset, value, REGISTER_ACCESS_PASSWORD)



import json

from qcodes.utils.helpers import full_class, named_repr


def test_full_class():
    j = json.JSONEncoder()
    assert full_class(j) == 'json.encoder.JSONEncoder'


def test_named_repr():
    j = json.JSONEncoder()
    id_ = id(j)
    j.name = 'Peppa'
    assert named_repr(j) == f'<json.encoder.JSONEncoder: Peppa at {id_}>'



import pytest
from qcodes.utils.validators import Enum, Ints, Lists, MultiType, Strings

from .conftest import AClass, a_func


def test_good():
    m = MultiType(Strings(2, 4), Ints(10, 1000))

    for v in [10, 11, 123, 1000, 'aa', 'mop', 'FRED']:
        m.validate(v)

    for v in [9, 1001, 'Q', 'Qcode', None, 100.0, b'nice', [], {},
              a_func, AClass, AClass(), True, False]:
        with pytest.raises(ValueError):
            m.validate(v)

    assert repr(m) == '<MultiType: Strings 2<=len<=4, Ints 10<=v<=1000>'

    assert m.is_numeric

    assert not MultiType(Strings(), Enum(1, 2)).is_numeric


def test_bad():
    for args in [[], [1], [Strings(), True]]:
        with pytest.raises(TypeError):
            MultiType(*args)


def test_valid_values():
    ms = [MultiType(Strings(2, 4), Ints(10, 32)),
          MultiType(Ints(), Lists(Ints())),
          MultiType(Ints(), MultiType(Lists(Ints()), Ints()))]
    for m in ms:
        for val in m.valid_values:
            m.validate(val)



import pytest
import logging
import io

from qcodes.instrument_drivers.stahl import Stahl
import qcodes.instrument.sims as sims


@pytest.fixture(scope="function")
def stahl_instrument():
    visa_lib = sims.__file__.replace(
        '__init__.py',
        'stahl.yaml@sim'
    )

    inst = Stahl('Stahl', 'ASRL3', visalib=visa_lib)
    inst.log.setLevel(logging.DEBUG)
    iostream = io.StringIO()
    lh = logging.StreamHandler(iostream)
    inst.log.logger.addHandler(lh)

    try:
        yield inst
    finally:
        inst.close()


def test_parse_idn_string():
    """
    Test that we can parse IDN strings correctly
    """
    assert Stahl.parse_idn_string("HV123 005 16 b") == {
        "model": "HV",
        "serial_number": "123",
        "voltage_range": 5.0,
        "n_channels": 16,
        "output_type": "bipolar"
    }

    with pytest.raises(
            RuntimeError,
            match="Unexpected instrument response"
    ):
        Stahl.parse_idn_string("HS123 005 16 bla b")


def test_get_idn(stahl_instrument):
    """
    Instrument attributes are set correctly after getting the IDN
    """
    assert stahl_instrument.IDN() == {
        "vendor": "Stahl",
        "model": "BS",
        "serial": "123",
        "firmware": None
    }

    assert stahl_instrument.n_channels == 16
    assert stahl_instrument.voltage_range == 5.0
    assert stahl_instrument.output_type == "bipolar"


def test_get_set_voltage(stahl_instrument):
    """
    Test that we can correctly get/set voltages
    """
    stahl_instrument.channel[0].voltage(1.2)
    assert stahl_instrument.channel[0].voltage() == -1.2
    logger = stahl_instrument.log.logger
    log_messages = logger.handlers[0].stream.getvalue()
    assert "did not produce an acknowledge reply" not in log_messages


def test_get_set_voltage_assert_warning(stahl_instrument):
    """
    On channel 2 we have deliberately introduced an error in the
    visa simulation; setting a voltage does not produce an acknowledge
    string. Test that a warning is correctly issued.
    """
    stahl_instrument.channel[1].voltage(1.0)
    logger = stahl_instrument.log.logger
    log_messages = logger.handlers[0].stream.getvalue()
    assert "did not produce an acknowledge reply" in log_messages


def test_get_current(stahl_instrument):
    """
    Test that we can read currents and that the unit is in Ampere
    """
    assert stahl_instrument.channel[0].current() == 1E-6
    assert stahl_instrument.channel[0].current.unit == "A"


def test_get_temperature(stahl_instrument):
    """
    Due to limitations in pyvisa-sim, we cannot test this.
    Line 191 of pyvisa-sim/component.py  should read
    "return response.encode('latin-1')" for this to work.
    """
    pass



import time
from datetime import datetime

import pytest
from qcodes.logger.logger import LogCapture
from qcodes.utils.helpers import wait_secs


def test_bad_calls():
    bad_args = [None, datetime.now()]
    for arg in bad_args:
        with pytest.raises(TypeError):
            wait_secs(arg)


def test_warning():
    with LogCapture() as logs:
        secs_out = wait_secs(time.perf_counter() - 1)
    assert secs_out == 0

    assert logs.value.count('negative delay') == 1, logs.value



import math
from typing import Any, List

import numpy as np
import pytest
from qcodes.utils.validators import Numbers

from .conftest import AClass, a_func

numbers = [0, 1, -1, 0.1, -0.1, 100, 1000000, 1.0, 3.5, -2.3e6, 5.5e15,
           1.34e-10, -2.5e-5, math.pi, math.e,
           # warning: True==1 and False==0
           True, False,
           # warning: +/- inf are allowed if max & min are not specified!
           -float("inf"), float("inf"),
           # numpy scalars
           np.int64(36), np.float32(-1.123)
           ]
not_numbers: List[Any] = ['', None, '1', [], {}, [1, 2], {1: 1},
                          b'good', AClass, AClass(), a_func]


def test_unlimited():
    n = Numbers()

    for v in numbers:
        n.validate(v)

    for v in not_numbers:
        with pytest.raises(TypeError):
            n.validate(v)

    # special case - nan now raises a ValueError rather than a TypeError
    with pytest.raises(ValueError):
        n.validate(float('nan'))

    n.validate(n.valid_values[0])


def test_min():
    for min_val in [-1e20, -1, -0.1, 0, 0.1, 10]:
        n = Numbers(min_value=min_val)

        n.validate(n.valid_values[0])
        for v in numbers:
            if v >= min_val:
                n.validate(v)
            else:
                with pytest.raises(ValueError):
                    n.validate(v)

    for v in not_numbers:
        with pytest.raises(TypeError):
            n.validate(v)

    with pytest.raises(ValueError):
        n.validate(float('nan'))


def test_max():
    for max_val in [-1e20, -1, -0.1, 0, 0.1, 10]:
        n = Numbers(max_value=max_val)

        n.validate(n.valid_values[0])
        for v in numbers:
            if v <= max_val:
                n.validate(v)
            else:
                with pytest.raises(ValueError):
                    n.validate(v)

    for v in not_numbers:
        with pytest.raises(TypeError):
            n.validate(v)

    with pytest.raises(ValueError):
        n.validate(float('nan'))


def test_range():
    n = Numbers(0.1, 3.5)

    for v in numbers:
        if 0.1 <= v <= 3.5:
            n.validate(v)
        else:
            with pytest.raises(ValueError):
                n.validate(v)

    for v in not_numbers:
        with pytest.raises(TypeError):
            n.validate(v)

    with pytest.raises(ValueError):
        n.validate(float('nan'))

    assert repr(n) == '<Numbers 0.1<=v<=3.5>'


def test_failed_numbers():
    with pytest.raises(TypeError):
        Numbers(1, 2, 3)

    with pytest.raises(TypeError):
        Numbers(1, 1)  # min >= max

    for val in not_numbers:
        with pytest.raises(TypeError):
            Numbers(max_value=val)

        with pytest.raises(TypeError):
            Numbers(min_value=val)


def test_valid_values():
    val = Numbers()
    for vval in val.valid_values:
        val.validate(vval)



import math

import numpy as np
import pytest
from qcodes.utils.validators import Bool

from .conftest import AClass, a_func

BOOLS = [True, False, np.bool8(True), np.bool8(False)]
NOTBOOLS = [0, 1, 10, -1, 100, 1000000, int(-1e15), int(1e15),
            0.1, -0.1, 1.0, 3.5, -2.3e6, 5.5e15, 1.34e-10, -2.5e-5,
            math.pi, math.e, '', None, float("nan"), float("inf"),
            -float("inf"), '1', [], {}, [1, 2], {1: 1}, b'good',
            AClass, AClass(), a_func]


def test_bool():
    b = Bool()

    for v in BOOLS:
        b.validate(v)

    for v in NOTBOOLS:
        with pytest.raises(TypeError):
            b.validate(v)

        assert repr(b) == '<Boolean>'


def test_valid_bool_values():
    val = Bool()
    for vval in val.valid_values:
        val.validate(vval)



from unittest import TestCase
import numpy as np

import qcodes
import qcodes.measure
from qcodes.data.hdf5_format import HDF5Format, HDF5FormatMetadata
from qcodes.data.gnuplot_format import GNUPlotFormat
from qcodes.data.data_set import load_data
from qcodes.tests.legacy.data_mocks import DataSet2D


#%%
class TestFormatters(TestCase):

    def setUp(self):
        self.formatters = [GNUPlotFormat, HDF5Format, HDF5FormatMetadata]
        self.metadata = {'subdict': {'stringlist': ['P1']}, 'string': 'P1',
                         'int': 1, 'list': [1, 2], 'numpyarray': np.array([1])}

    def test_read_write(self):
        for f in self.formatters:
            print('test formatter %s' % f)
            dataset = DataSet2D(name="test_read_write")
            dataset.formatter = f()

            dataset.add_metadata(self.metadata)
            dataset.write(write_metadata=True)

            dataset2 = load_data(dataset.location, formatter=f())
            self.assertEqual(list(dataset.arrays.keys()),
                             list(dataset2.arrays.keys()))
            # strings should be read and written identically
            self.assertEqual(dataset.metadata['string'],
                             dataset2.metadata['string'])


class TestNoSorting(TestCase):
    """
    (WilliamHPNielsen): I'm not too sure where this test belongs... It tests
    that parameters with non-sortable keys can be saved using the gnuplot
    formatter, so I guess it goes here.
    """

    param = qcodes.Parameter(name='mixed_val_mapping_param',
                             get_cmd=lambda: np.random.randint(1, 3),
                             val_mapping={1: 1, '2': 2}
                             )

    def test_can_measure(self):
        qcodes.measure.Measure(self.param).run(name="test_no_sorting")



import re
from unittest.mock import MagicMock, call
import numpy as np
import pytest

from qcodes.instrument_drivers.Keysight.keysightb1500 import constants
from qcodes.instrument_drivers.Keysight.keysightb1500.KeysightB1520A import \
    B1520A


# pylint: disable=redefined-outer-name

@pytest.fixture
def mainframe():
    yield MagicMock()


@pytest.fixture
def cmu(mainframe):
    slot_nr = 3
    cmu = B1520A(parent=mainframe, name='B1520A', slot_nr=slot_nr)
    # GroupParameter with initial values write at the init so reset the mock
    # to not count those write
    mainframe.reset_mock()

    yield cmu


def test_force_dc_voltage(cmu):
    mainframe = cmu.parent

    cmu.voltage_dc(10)

    mainframe.write.assert_called_once_with('DCV 3,10')


def test_force_ac_voltage(cmu):
    mainframe = cmu.parent

    cmu.voltage_ac(0.1)

    mainframe.write.assert_called_once_with('ACV 3,0.1')


def test_set_ac_frequency(cmu):
    mainframe = cmu.parent

    cmu.frequency(100e3)

    mainframe.write.assert_called_once_with('FC 3,100000.0')


def test_get_dc_voltage(cmu):
    mainframe = cmu.parent
    mainframe.ask.return_value = 'DCV3,0.000;ACV3,0.0;FC3,1000.000'
    response = cmu.voltage_dc()
    assert response == 0.0

    mainframe.reset_mock()
    mainframe.ask.return_value = 'DCV3,0.001;ACV3,0.0;FC3,1000.000'
    response = cmu.voltage_dc()
    assert response == 0.001

    mainframe.reset_mock()
    mainframe.ask.return_value = 'DCV3,13.051;ACV3,0.0;FC3,1000.000'
    response = cmu.voltage_dc()
    assert response == 13.051


def test_get_ac_voltage(cmu):
    mainframe = cmu.parent
    mainframe.ask.return_value = 'DCV3,0.000;ACV3,0.000;FC3,1000.000'
    response = cmu.voltage_ac()
    assert response == 0.0

    mainframe.reset_mock()
    mainframe.ask.return_value = 'DCV3,0.001;ACV3,0.01;FC3,1000.000'
    response = cmu.voltage_ac()
    assert response == 0.01

    mainframe.reset_mock()
    mainframe.ask.return_value = 'DCV3,13.051;ACV3,3.045;FC3,1000.000'
    response = cmu.voltage_ac()
    assert response == 3.045


def test_get_frequency(cmu):
    mainframe = cmu.parent
    mainframe.ask.return_value = 'DCV3,0.000;ACV3,0.000;FC3,100000.000'
    response = cmu.frequency()
    assert response == 100000.0

    mainframe.reset_mock()
    mainframe.ask.return_value = 'DCV3,0.001;ACV3,0.01;FC3,1000.033'
    response = cmu.frequency()
    assert response == 1000.033

    mainframe.reset_mock()
    mainframe.ask.return_value = 'DCV3,13.051;ACV3,3.045;FC3,1540.40'
    response = cmu.frequency()
    assert response == 1540.40


def test_get_capacitance(cmu):
    mainframe = cmu.parent

    mainframe.ask.return_value = "NCC-1.45713E-06,NCD-3.05845E-03"

    assert pytest.approx((-1.45713E-06, -3.05845E-03)) == cmu.capacitance()

    mainframe.ask.return_value = "NCC-1.55713E-06,NCD-3.15845E-03," \
                                 "NCV-1.52342E-03,NCV+0.14235E-03"

    assert pytest.approx((-1.55713E-06, -3.15845E-03)) == cmu.capacitance()


def test_raise_error_on_unsupported_result_format(cmu):
    mainframe = cmu.parent

    mainframe.ask.return_value = "NCR-1.1234E-03,NCX-4.5677E-03,NCV+0.14235E-03"

    with pytest.raises(ValueError):
        cmu.capacitance()


def test_ranging_mode(cmu):
    mainframe = cmu.parent

    cmu.ranging_mode(constants.RangingMode.AUTO)

    mainframe.write.assert_called_once_with('RC 3,0')


def test_set_sweep_auto_abort(cmu):
    mainframe = cmu.parent

    cmu.cv_sweep.sweep_auto_abort(constants.Abort.ENABLED)

    mainframe.write.assert_called_once_with("WMDCV 2")


def test_get_sweep_auto_abort(cmu):
    mainframe = cmu.parent

    mainframe.ask.return_value = "WMDCV2,2;WTDCV1.0,0.0,0.0,0.0,0.0"
    condition = cmu.cv_sweep.sweep_auto_abort()
    assert condition == constants.Abort.ENABLED


def test_set_post_sweep_voltage_cond(cmu):
    mainframe = cmu.parent
    mainframe.ask.return_value = "WMDCV2,2;WTDCV1.0,0.0,0.0,0.0,0.0"
    cmu.cv_sweep.post_sweep_voltage_condition.set(constants.WMDCV.Post.STOP)

    mainframe.write.assert_called_once_with("WMDCV 2,2")


def test_get_post_sweep_voltage_cond(cmu):
    mainframe = cmu.parent

    mainframe.ask.return_value = "WMDCV2,2;WTDCV1.0,0.0,0.0,0.0,0.0"
    condition = cmu.cv_sweep.post_sweep_voltage_condition()
    assert condition == constants.WMDCV.Post.STOP

    mainframe.reset_mock()

    mainframe.ask.return_value = "WMDCV2;WTDCV1.0,0.0,0.0,0.0,0.0"
    msg = re.escape("Received None. Set the parameter"
                    "``post_sweep_voltage_condition`` first.")
    with pytest.raises(ValueError, match=msg):
        cmu.cv_sweep.post_sweep_voltage_condition()


def test_cv_sweep_delay(cmu):
    mainframe = cmu.root_instrument

    mainframe.ask.return_value = "WTDCV0.0,0.0,0.0,0.0,0.0"

    cmu.cv_sweep.hold_time(1.0)
    cmu.cv_sweep.delay(1.0)

    mainframe.write.assert_has_calls([call("WTDCV 1.0,0.0,0.0,0.0,0.0"),
                                      call("WTDCV 1.0,1.0,0.0,0.0,0.0")])


def test_cmu_sweep_steps(cmu):
    mainframe = cmu.root_instrument
    mainframe.ask.return_value = "WDCV3,1,0.0,0.0,1"
    cmu.cv_sweep.sweep_start(2.0)
    cmu.cv_sweep.sweep_end(4.0)

    mainframe.write.assert_has_calls([call("WDCV 3,1,2.0,0.0,1"),
                                      call("WDCV 3,1,2.0,4.0,1")])


def test_cv_sweep_voltages(cmu):

    mainframe = cmu.root_instrument

    start = -1.0
    end = 1.0
    steps = 5
    return_string = f'WDCV3,1,{start},{end},{steps}'
    mainframe.ask.return_value = return_string

    cmu.cv_sweep.sweep_start(start)
    cmu.cv_sweep.sweep_end(end)
    cmu.cv_sweep.sweep_steps(steps)
    voltages = cmu.cv_sweep_voltages()

    assert all([a == b for a, b in zip(np.linspace(start, end, steps),
                                       voltages)])


def test_sweep_modes(cmu):

    mainframe = cmu.root_instrument

    start = -1.0
    end = 1.0
    steps = 5
    mode = constants.SweepMode.LINEAR_TWO_WAY
    return_string = f'WDCV3,{mode},{start},{end},{steps}'
    mainframe.ask.return_value = return_string

    cmu.cv_sweep.sweep_start(start)
    cmu.cv_sweep.sweep_end(end)
    cmu.cv_sweep.sweep_steps(steps)
    cmu.cv_sweep.sweep_mode(mode)
    voltages = cmu.cv_sweep_voltages()

    assert all([a == b for a, b in zip((-1.0, 0.0, 1.0, 0.0, -1.0), voltages)])


def test_run_sweep(cmu):
    mainframe = cmu.root_instrument

    start = -1.0
    end = 1.0
    steps = 5

    return_string = f'WMDCV2,2;WTDCV0.00,0.0000,0.2250,0.0000,' \
                    f'0.0000;WDCV3,' \
                    f'1,{start},{end},{steps};ACT0,1'
    mainframe.ask.return_value = return_string
    cmu.setup_fnc_already_run = True
    cmu.impedance_model(constants.IMP.MeasurementMode.G_X)
    cmu.cv_sweep.sweep_start(start)
    cmu.cv_sweep.sweep_end(end)
    cmu.cv_sweep.sweep_steps(steps)
    cmu.adc_mode(constants.ACT.Mode.PLC)
    cmu.adc_coef(5)
    cmu.run_sweep()
    mainframe.write.assert_has_calls([
        call('WDCV 3,1,-1.0,0.0,1'),
        call('WDCV 3,1,-1.0,1.0,1'),
        call('WDCV 3,1,-1.0,1.0,5'),
        call('ACT 2,1'),
        call('ACT 2,5')])
    assert cmu.run_sweep.names == ('conductance', 'reactance')
    assert cmu.run_sweep.labels == ('Conductance', 'Reactance')
    assert cmu.run_sweep.units == ('S', 'ohms')




def test_phase_compensation_mode(cmu):
    mainframe = cmu.parent

    cmu.phase_compensation_mode(constants.ADJ.Mode.MANUAL)

    mainframe.write.assert_called_once_with('ADJ 3,1')

    assert constants.ADJ.Mode.MANUAL == cmu.phase_compensation_mode()


def test_phase_compensation(cmu):
    mainframe = cmu.parent

    mainframe.ask.return_value = 0

    response = cmu.phase_compensation()

    mainframe.ask.assert_called_once_with('ADJ? 3')
    assert isinstance(response, constants.ADJQuery.Response)
    assert response == constants.ADJQuery.Response.PASSED


def test_phase_compensation_with_mode(cmu):
    mainframe = cmu.parent

    mainframe.ask.return_value = 0

    response = cmu.phase_compensation(constants.ADJQuery.Mode.USE_LAST)

    mainframe.ask.assert_called_once_with('ADJ? 3,0')
    assert isinstance(response, constants.ADJQuery.Response)
    assert response == constants.ADJQuery.Response.PASSED


def test_enable_correction(cmu):
    mainframe = cmu.parent

    cmu.correction.enable(constants.CalibrationType.OPEN)
    mainframe.write.assert_called_once_with('CORRST 3,1,1')

    mainframe.reset_mock()

    cmu.correction.enable(constants.CalibrationType.SHORT)
    mainframe.write.assert_called_once_with('CORRST 3,2,1')

    mainframe.reset_mock()

    cmu.correction.enable(constants.CalibrationType.LOAD)
    mainframe.write.assert_called_once_with('CORRST 3,3,1')


def test_disable_correction(cmu):
    mainframe = cmu.parent

    cmu.correction.disable(constants.CalibrationType.OPEN)
    mainframe.write.assert_called_once_with('CORRST 3,1,0')

    mainframe.reset_mock()

    cmu.correction.disable(constants.CalibrationType.SHORT)
    mainframe.write.assert_called_once_with('CORRST 3,2,0')

    mainframe.reset_mock()

    cmu.correction.disable(constants.CalibrationType.LOAD)
    mainframe.write.assert_called_once_with('CORRST 3,3,0')


def test_correction_is_enabled(cmu):
    mainframe = cmu.parent

    mainframe.ask.return_value = '1'

    response = cmu.correction.is_enabled(constants.CalibrationType.SHORT)
    assert response == constants.CORRST.Response.ON


def test_correction_set_reference_values(cmu):
    mainframe = cmu.parent

    cmu.correction.set_reference_values(
        constants.CalibrationType.OPEN,
        constants.DCORR.Mode.Cp_G,
        1,
        2)
    mainframe.write.assert_called_once_with('DCORR 3,1,100,1,2')


def test_correction_get_reference_values(cmu):
    mainframe = cmu.parent

    mainframe.ask.return_value = '100,0.001,2'
    response = 'Mode: Cp_G, Primary Cp: 0.001 F, Secondary G: 2.0 S'
    assert response == cmu.correction.get_reference_values(
        constants.CalibrationType.OPEN)


def test_clear_and_set_default_frequency_list_for_correction(cmu):
    mainframe = cmu.parent

    cmu.correction.frequency_list.clear_and_set_default()

    mainframe.write.assert_called_once_with('CLCORR 3,2')


def test_clear_frequency_list_for_correction(cmu):
    mainframe = cmu.parent

    cmu.correction.frequency_list.clear()

    mainframe.write.assert_called_once_with('CLCORR 3,1')


def test_add_frequency_for_correction(cmu):
    mainframe = cmu.parent

    cmu.correction.frequency_list.add(1000)

    mainframe.write.assert_called_once_with('CORRL 3,1000')


def test_query_from_frequency_list_for_correction(cmu):
    mainframe = cmu.parent

    mainframe.ask.return_value = '25'

    assert pytest.approx(25) == cmu.correction.frequency_list.query()
    mainframe.ask.assert_called_once_with('CORRL? 3')


def test_query_at_index_from_frequency_list_for_correction(cmu):
    mainframe = cmu.parent

    mainframe.ask.return_value = '1234.567'

    assert pytest.approx(1234.567) == cmu.correction.frequency_list.query(
        index=0)
    mainframe.ask.assert_called_once_with('CORRL? 3,0')


def test_perform_correction(cmu):
    mainframe = cmu.parent

    mainframe.ask.return_value = 0

    response = cmu.correction.perform(
        constants.CalibrationType.OPEN)
    assert constants.CORR.Response.SUCCESSFUL == response


def test_perform_and_enable_correction(cmu):
    mainframe = cmu.parent

    mainframe.ask.side_effect = [
        '0',  # for correction status
        '1'   # for correction state (enabled/disabled)
    ]

    response = cmu.correction.perform_and_enable(
        constants.CalibrationType.OPEN)

    expected_response = f'Correction status ' \
                        f'{constants.CORR.Response.SUCCESSFUL.name} and ' \
                        f'Enable ' \
                        f'{constants.CORRST.Response.ON.name}'
    assert response == expected_response


def test_abort(cmu):
    mainframe = cmu.parent

    cmu.abort()

    mainframe.write.assert_called_once_with('AB')



"""
This plotting module provides various functions to plot the data measured
using QCoDeS.
"""

import inspect
import logging
from contextlib import contextmanager
from functools import partial
from typing import Any, List, Optional, Sequence, Tuple, Union, cast

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import FuncFormatter

import qcodes as qc
from qcodes.dataset.data_set import load_by_run_spec
from qcodes.dataset.data_set_protocol import DataSetProtocol
from qcodes.utils.plotting import auto_color_scale_from_config, find_scale_and_prefix

from .data_export import (
    DSPlotData,
    _get_data_from_ds,
    _strings_as_ints,
    get_1D_plottype,
    get_2D_plottype,
    reshape_2D_data,
)

log = logging.getLogger(__name__)
DB = qc.config["core"]["db_location"]

AxesTuple = Tuple[matplotlib.axes.Axes, matplotlib.colorbar.Colorbar]
AxesTupleList = Tuple[List[matplotlib.axes.Axes],
                      List[Optional[matplotlib.colorbar.Colorbar]]]
# NamedData is the structure _get_data_from_ds returns and that plot_by_id
# uses internally
NamedData = List[List[DSPlotData]]

# list of kwargs for plotting function, so that kwargs can be passed to
# :func:`plot_dataset` and will be distributed to the respective plotting func.
# subplots passes on the kwargs called `fig_kw` to the underlying `figure` call
# First find the kwargs that belong to subplots and than add those that are
# redirected to the `figure`-call.
SUBPLOTS_OWN_KWARGS = set(inspect.signature(plt.subplots).parameters.keys())
SUBPLOTS_OWN_KWARGS.remove('fig_kw')
FIGURE_KWARGS = set(inspect.signature(plt.figure).parameters.keys())
FIGURE_KWARGS.remove('kwargs')
SUBPLOTS_KWARGS = SUBPLOTS_OWN_KWARGS.union(FIGURE_KWARGS)


@contextmanager
def _appropriate_kwargs(plottype: str,
                        colorbar_present: bool,
                        **kwargs: Any) -> Any:
    """
    NB: Only to be used inside :func"`plot_dataset`.

    Context manager to temporarily mutate the plotting kwargs to be appropriate
    for a specific plottype. This is helpful since :func:`plot_dataset` may have
    to generate different kinds of plots (e.g. heatmaps and line plots) and
    the user may want to specify kwargs only relevant to some of them
    (e.g. 'cmap', that line plots cannot consume). Those kwargs should then not
    be passed to all plots, which is what this contextmanager handles.

    Args:
        plottype: The plot type for which the kwargs should be adjusted
        colorbar_present: Is there a non-None colorbar in this plot iteration?
    """

    def linehandler(**kwargs: Any) -> Any:
        kwargs.pop('cmap', None)
        return kwargs

    def heatmaphandler(**kwargs: Any) -> Any:
        if not(colorbar_present) and 'cmap' not in kwargs:
            kwargs['cmap'] = qc.config.plotting.default_color_map
        return kwargs

    plot_handler_mapping = {'1D_line': linehandler,
                            '1D_point': linehandler,
                            '1D_bar': linehandler,
                            '2D_point': heatmaphandler,
                            '2D_grid': heatmaphandler,
                            '2D_scatter': heatmaphandler,
                            '2D_equidistant': heatmaphandler,
                            '2D_unknown': heatmaphandler}

    yield plot_handler_mapping[plottype](**kwargs.copy())


def plot_dataset(
    dataset: DataSetProtocol,
    axes: Optional[Union[matplotlib.axes.Axes, Sequence[matplotlib.axes.Axes]]] = None,
    colorbars: Optional[
        Union[matplotlib.colorbar.Colorbar, Sequence[matplotlib.colorbar.Colorbar]]
    ] = None,
    rescale_axes: bool = True,
    auto_color_scale: Optional[bool] = None,
    cutoff_percentile: Optional[Union[Tuple[float, float], float]] = None,
    complex_plot_type: str = "real_and_imag",
    complex_plot_phase: str = "radians",
    **kwargs: Any,
) -> AxesTupleList:
    """
    Construct all plots for a given dataset

    Implemented so far:
       * 1D line and scatter plots
       * 2D plots on filled out rectangular grids
       * 2D scatterplots (fallback)

    The function can optionally be supplied with a matplotlib axes or a list
    of axes that will be used for plotting. The user should ensure that the
    number of axes matches the number of datasets to plot. To plot several (1D)
    dataset in the same axes supply it several times. Colorbar axes are
    created dynamically. If colorbar axes are supplied, they will be reused,
    yet new colorbar axes will be returned.

    The plot has a title that comprises run id, experiment name, and sample
    name.

    ``**kwargs`` are passed to matplotlib's relevant plotting functions
    By default the data in any vector plot will be rasterized
    for scatter plots and heatmaps if more than 5000 points are supplied.
    This can be overridden by supplying the `rasterized` kwarg.

    Args:
        dataset: The dataset to plot
        axes: Optional Matplotlib axes to plot on. If not provided, new axes
            will be created
        colorbars: Optional Matplotlib Colorbars to use for 2D plots. If not
            provided, new ones will be created
        rescale_axes: If True, tick labels and units for axes of parameters
            with standard SI units will be rescaled so that, for example,
            '0.00000005' tick label on 'V' axis are transformed to '50' on 'nV'
            axis ('n' is 'nano')
        auto_color_scale: If True, the colorscale of heatmap plots will be
            automatically adjusted to disregard outliers.
        cutoff_percentile: Percentile of data that may maximally be clipped
            on both sides of the distribution.
            If given a tuple (a,b) the percentile limits will be a and 100-b.
            See also the plotting tuorial notebook.
        complex_plot_type: Method for converting complex-valued parameters
            into two real-valued parameters, either ``"real_and_imag"`` or
            ``"mag_and_phase"``. Applicable only for the cases where the
            dataset contains complex numbers
        complex_plot_phase: Format of phase for plotting complex-valued data,
            either ``"radians"`` or ``"degrees"``. Applicable only for the
            cases where the dataset contains complex numbers

    Returns:
        A list of axes and a list of colorbars of the same length. The
        colorbar axes may be `None` if no colorbar is created (e.g. for
        1D plots)

    Config dependencies: (qcodesrc.json)
    """

    # handle arguments and defaults
    subplots_kwargs = {k: kwargs.pop(k)
                       for k in set(kwargs).intersection(SUBPLOTS_KWARGS)}

    # sanitize the complex plotting kwargs
    if complex_plot_type not in ['real_and_imag', 'mag_and_phase']:
        raise ValueError(
            f'Invalid complex plot type given. Received {complex_plot_type} '
            'but can only accept "real_and_imag" or "mag_and_phase".')
    if complex_plot_phase not in ['radians', 'degrees']:
        raise ValueError(
            f'Invalid complex plot phase given. Received {complex_plot_phase} '
            'but can only accept "degrees" or "radians".')
    degrees = complex_plot_phase == "degrees"

    # Retrieve info about the run for the title

    experiment_name = dataset.exp_name
    sample_name = dataset.sample_name
    title = f"Run #{dataset.captured_run_id}, " \
            f"Experiment {experiment_name} ({sample_name})"

    alldata: NamedData = _get_data_from_ds(dataset)
    alldata = _complex_to_real_preparser(alldata,
                                         conversion=complex_plot_type,
                                         degrees=degrees)

    nplots = len(alldata)

    if isinstance(axes, matplotlib.axes.Axes):
        axeslist = [axes]
    else:
        axeslist = cast(List[matplotlib.axes.Axes], axes)
    if isinstance(colorbars, matplotlib.colorbar.Colorbar):
        colorbars = [colorbars]

    if axeslist is None:
        axeslist = []
        for i in range(nplots):
            fig, ax = plt.subplots(1, 1, **subplots_kwargs)
            axeslist.append(ax)
    else:
        if len(subplots_kwargs) != 0:
            raise RuntimeError(f"Error: You cannot provide arguments for the "
                               f"axes/figure creation if you supply your own "
                               f"axes. "
                               f"Provided arguments: {subplots_kwargs}")
        if len(axeslist) != nplots:
            raise RuntimeError(f"Trying to make {nplots} plots, but"
                               f"received {len(axeslist)} axes objects.")

    if colorbars is None:
        colorbars = len(axeslist)*[None]
    new_colorbars: List[matplotlib.colorbar.Colorbar] = []

    for data, ax, colorbar in zip(alldata, axeslist, colorbars):

        if len(data) == 2:  # 1D PLOTTING
            log.debug(f'Doing a 1D plot with kwargs: {kwargs}')

            xpoints = data[0]["data"]
            ypoints = data[1]["data"]

            plottype = get_1D_plottype(xpoints, ypoints)
            log.debug(f'Determined plottype: {plottype}')

            if plottype == '1D_line':
                # sort for plotting
                order = xpoints.argsort()
                xpoints = xpoints[order]
                ypoints = ypoints[order]

                with _appropriate_kwargs(plottype,
                                         colorbar is not None, **kwargs) as k:
                    ax.plot(xpoints, ypoints, **k)
            elif plottype == '1D_point':
                with _appropriate_kwargs(plottype,
                                         colorbar is not None, **kwargs) as k:
                    ax.scatter(xpoints, ypoints, **k)
            elif plottype == '1D_bar':
                with _appropriate_kwargs(plottype,
                                         colorbar is not None, **kwargs) as k:
                    ax.bar(xpoints, ypoints, **k)
            else:
                raise ValueError('Unknown plottype. Something is way wrong.')

            _set_data_axes_labels(ax, data)

            if rescale_axes:
                _rescale_ticks_and_units(ax, data, colorbar)

            new_colorbars.append(None)

            ax.set_title(title)

        elif len(data) == 3:  # 2D PLOTTING
            log.debug(f'Doing a 2D plot with kwargs: {kwargs}')

            if data[2]["shape"] is None:
                xpoints = data[0]["data"].flatten()
                ypoints = data[1]["data"].flatten()
                zpoints = data[2]["data"].flatten()
                plottype = get_2D_plottype(xpoints, ypoints, zpoints)
                log.debug(f"Determined plottype: {plottype}")
            else:
                xpoints = data[0]["data"]
                ypoints = data[1]["data"]
                zpoints = data[2]["data"]
                plottype = "2D_grid"

            how_to_plot = {'2D_grid': plot_on_a_plain_grid,
                           '2D_equidistant': plot_on_a_plain_grid,
                           '2D_point': plot_2d_scatterplot,
                           '2D_unknown': plot_2d_scatterplot}
            plot_func = how_to_plot[plottype]

            with _appropriate_kwargs(plottype,
                                     colorbar is not None, **kwargs) as k:
                ax, colorbar = plot_func(xpoints, ypoints, zpoints,
                                         ax, colorbar,
                                         **k)

            _set_data_axes_labels(ax, data, colorbar)

            if rescale_axes:
                _rescale_ticks_and_units(ax, data, colorbar)

            auto_color_scale_from_config(colorbar, auto_color_scale,
                                         zpoints, cutoff_percentile)

            new_colorbars.append(colorbar)

            ax.set_title(title)

        else:
            log.warning('Multi-dimensional data encountered. '
                        f'parameter {data[-1]["name"]} depends on '
                        f'{len(data)-1} parameters, cannot plot '
                        f'that.')
            new_colorbars.append(None)

    if len(axeslist) != len(new_colorbars):
        raise RuntimeError("Non equal number of axes. Perhaps colorbar is "
                           "missing from one of the cases above")
    return axeslist, new_colorbars


def plot_by_id(
    run_id: int,
    axes: Optional[Union[matplotlib.axes.Axes, Sequence[matplotlib.axes.Axes]]] = None,
    colorbars: Optional[
        Union[matplotlib.colorbar.Colorbar, Sequence[matplotlib.colorbar.Colorbar]]
    ] = None,
    rescale_axes: bool = True,
    auto_color_scale: Optional[bool] = None,
    cutoff_percentile: Optional[Union[Tuple[float, float], float]] = None,
    complex_plot_type: str = "real_and_imag",
    complex_plot_phase: str = "radians",
    **kwargs: Any,
) -> AxesTupleList:
    """
    Construct all plots for a given `run_id`. Here `run_id` is an
    alias for `captured_run_id` for historical reasons. See the docs
    of :func:`.load_by_run_spec` for details of loading runs.
    All other arguments are forwarded
    to :func:`.plot_dataset`, see this for more details.
    """

    dataset = load_by_run_spec(captured_run_id=run_id)
    return plot_dataset(dataset,
                        axes,
                        colorbars,
                        rescale_axes,
                        auto_color_scale,
                        cutoff_percentile,
                        complex_plot_type,
                        complex_plot_phase,
                        **kwargs)


def _complex_to_real_preparser(
    alldata: Sequence[Sequence[DSPlotData]], conversion: str, degrees: bool = False
) -> NamedData:
    """
    Convert complex-valued parameters to two real-valued parameters, either
    real and imaginary part or phase and magnitude part

    Args:
        alldata: The data to convert, should be the output of `_get_data_from_ds`
        conversion: the conversion method, either "real_and_imag" or
            "mag_and_phase"
        degrees: Whether to return the phase in degrees. The default is to
            return the phase in radians
    """

    if conversion not in ['real_and_imag', 'mag_and_phase']:
        raise ValueError(f'Invalid conversion given. Received {conversion}, '
                         'but can only accept "real_and_imag" or '
                         '"mag_and_phase".')

    newdata: NamedData = []

    # we build a new NamedData object from the given `alldata` input.
    # Note that the length of `newdata` will be larger than that of `alldata`
    # in the case of complex top-level parameters, because a single complex
    # top-level parameter will be split into two real top-level parameters
    # (that have the same setpoints). This is the reason why we
    # use two variables below, new_group and new_groups.

    for group in alldata:
        new_group = []
        new_groups: NamedData = [[], []]
        for index, parameter in enumerate(group):
            data = parameter["data"]
            if data.dtype.kind == "c":
                p1, p2 = _convert_complex_to_real(
                    parameter, conversion=conversion, degrees=degrees
                )
                if index < len(group) - 1:
                    # if the above condition is met, we are dealing with
                    # complex setpoints
                    new_group.append(p1)
                    new_group.append(p2)
                else:
                    # in this case, we are dealing with a complex top-level
                    # parameter. Also, all the setpoints will have been handled
                    # by now. We split the group into two groups, one for each
                    # new (real) top-level parameter
                    new_groups[0] = new_group.copy()
                    new_groups[1] = new_group.copy()
                    new_groups[0].append(p1)
                    new_groups[1].append(p2)
            else:
                new_group.append(parameter)
        if new_groups == [[], []]:
            # if the above condition is met, the group did not contain a
            # complex top-level parameter and has thus not been split into two
            # new groups
            newdata.append(new_group)
        else:
            newdata.append(new_groups[0])
            newdata.append(new_groups[1])

    return newdata


def _convert_complex_to_real(
    parameter: DSPlotData, conversion: str, degrees: bool
) -> Tuple[DSPlotData, DSPlotData]:
    """
    Do the actual conversion and turn one parameter into two.
    Should only be called from within _complex_to_real_preparser.
    """

    phase_unit = 'deg' if degrees else 'rad'

    converters = {
        'data': {'real_and_imag': lambda x: (np.real(x), np.imag(x)),
                 'mag_and_phase': lambda x: (np.abs(x),
                                             np.angle(x, deg=degrees))},
        'labels': {'real_and_imag': lambda l: (l + ' [real]', l + ' [imag]'),
                   'mag_and_phase': lambda l: (l + ' [mag]', l + ' [phase]')},
        'units': {'real_and_imag': lambda u: (u, u),
                  'mag_and_phase': lambda u: (u, phase_unit)},
        'names': {'real_and_imag': lambda n: (n + '_real', n + '_imag'),
                  'mag_and_phase': lambda n: (n + '_mag', n + '_phase')}}

    new_data = converters['data'][conversion](parameter['data'])
    new_labels = converters['labels'][conversion](parameter['label'])
    new_units = converters['units'][conversion](parameter['unit'])
    new_names = converters['names'][conversion](parameter['name'])

    parameter1: DSPlotData = {
        "name": new_names[0],
        "label": new_labels[0],
        "unit": new_units[0],
        "data": new_data[0],
        "shape": parameter["shape"],
    }

    parameter2: DSPlotData = {
        "name": new_names[1],
        "label": new_labels[1],
        "unit": new_units[1],
        "data": new_data[1],
        "shape": parameter["shape"],
    }

    return parameter1, parameter2


def _get_label_of_data(data_dict: DSPlotData) -> str:
    return data_dict['label'] if data_dict['label'] != '' \
        else data_dict['name']


def _make_axis_label(label: str, unit: str) -> str:
    label = f'{label}'
    if unit != '' and unit is not None:
        label += f' ({unit})'
    return label


def _make_label_for_data_axis(data: Sequence[DSPlotData], axis_index: int) -> str:
    label = _get_label_of_data(data[axis_index])
    unit = data[axis_index]['unit']
    return _make_axis_label(label, unit)


def _set_data_axes_labels(
    ax: matplotlib.axes.Axes,
    data: Sequence[DSPlotData],
    cax: Optional[matplotlib.colorbar.Colorbar] = None,
) -> None:
    ax.set_xlabel(_make_label_for_data_axis(data, 0))
    ax.set_ylabel(_make_label_for_data_axis(data, 1))

    if cax is not None and len(data) > 2:
        cax.set_label(_make_label_for_data_axis(data, 2))


def plot_2d_scatterplot(x: np.ndarray, y: np.ndarray, z: np.ndarray,
                        ax: matplotlib.axes.Axes,
                        colorbar: matplotlib.colorbar.Colorbar = None,
                        **kwargs: Any) -> AxesTuple:
    """
    Make a 2D scatterplot of the data. ``**kwargs`` are passed to matplotlib's
    scatter used for the plotting. By default the data will be rasterized
    in any vector plot if more than 5000 points are supplied. This can be
    overridden by supplying the `rasterized` kwarg.

    Args:
        x: The x values
        y: The y values
        z: The z values
        ax: The axis to plot onto
        colorbar: The colorbar to plot into

    Returns:
        The matplotlib axis handles for plot and colorbar
    """
    if 'rasterized' in kwargs.keys():
        rasterized = kwargs.pop('rasterized')
    else:
        rasterized = len(z) > qc.config.plotting.rasterize_threshold

    z_is_stringy = isinstance(z[0], str)

    if z_is_stringy:
        z_strings = np.unique(z)
        z = _strings_as_ints(z)

    cmap = kwargs.pop('cmap') if 'cmap' in kwargs else None

    if z_is_stringy:
        name = cmap.name if hasattr(cmap, 'name') else 'viridis'
        cmap = matplotlib.cm.get_cmap(name, len(z_strings))

    mappable = ax.scatter(x=x, y=y, c=z,
                          rasterized=rasterized, cmap=cmap, **kwargs)

    if colorbar is not None:
        colorbar = ax.figure.colorbar(mappable, ax=ax, cax=colorbar.ax)
    else:
        colorbar = ax.figure.colorbar(mappable, ax=ax)

    if z_is_stringy:
        N = len(z_strings)
        f = (N-1)/N
        colorbar.set_ticks([(n+0.5)*f for n in range(N)])
        colorbar.set_ticklabels(z_strings)

    return ax, colorbar


def plot_on_a_plain_grid(x: np.ndarray,
                         y: np.ndarray,
                         z: np.ndarray,
                         ax: matplotlib.axes.Axes,
                         colorbar: matplotlib.colorbar.Colorbar = None,
                         **kwargs: Any
                         ) -> AxesTuple:
    """
    Plot a heatmap of z using x and y as axes. Assumes that the data
    are rectangular, i.e. that x and y together describe a rectangular
    grid. The arrays of x and y need not be sorted in any particular
    way, but data must belong together such that z[n] has x[n] and
    y[n] as setpoints.  The setpoints need not be equidistantly
    spaced, but linear interpolation is used to find the edges of the
    plotted squares. ``**kwargs`` are passed to matplotlib's pcolormesh used
    for the plotting. By default the data in any vector plot will be rasterized
    if more that 5000 points are supplied. This can be overridden
    by supplying the `rasterized` kwarg.

    Args:
        x: The x values
        y: The y values
        z: The z values
        ax: The axis to plot onto
        colorbar: A colorbar to reuse the axis for

    Returns:
        The matplotlib axes handle for plot and colorbar
    """

    log.debug(f'Got kwargs: {kwargs}')

    x_is_stringy = isinstance(x[0], str)
    y_is_stringy = isinstance(y[0], str)
    z_is_stringy = isinstance(z[0], str)

    if x_is_stringy:
        x_strings = np.unique(x)
        x = _strings_as_ints(x)

    if y_is_stringy:
        y_strings = np.unique(y)
        y = _strings_as_ints(y)

    if z_is_stringy:
        z_strings = np.unique(z)
        z = _strings_as_ints(z)

    if x.ndim == 2 and y.ndim == 2 and z.ndim == 2:
        if not np.logical_or(np.any(np.isnan(x)), np.any(np.isnan(y))):
            # data is on a grid that may or may not be
            # rectilinear. Rely on matplotlib to plot
            # this directly
            x_to_plot, y_to_plot, z_to_plot = x, y, z
            num_points = x_to_plot.size
        else:
            x_to_plot, y_to_plot, z_to_plot = _clip_nan_from_shaped_data(x, y, z)
            num_points = x_to_plot.size * y_to_plot.size
    else:
        x_to_plot, y_to_plot, z_to_plot = reshape_2D_data(x, y, z)
        num_points = x_to_plot.size * y_to_plot.size

    if 'rasterized' in kwargs.keys():
        rasterized = kwargs.pop('rasterized')
    else:
        rasterized = num_points > qc.config.plotting.rasterize_threshold

    cmap = kwargs.pop('cmap') if 'cmap' in kwargs else None

    if z_is_stringy:
        name = cmap.name if hasattr(cmap, 'name') else 'viridis'
        cmap = matplotlib.cm.get_cmap(name, len(z_strings))

    colormesh = ax.pcolormesh(
        x_to_plot,
        y_to_plot,
        np.ma.masked_invalid(z_to_plot),
        rasterized=rasterized,
        cmap=cmap,
        shading="nearest",
        **kwargs,
    )

    if x_is_stringy:
        ax.set_xticks(np.arange(len(np.unique(x_strings))))
        ax.set_xticklabels(x_strings)

    if y_is_stringy:
        ax.set_yticks(np.arange(len(np.unique(y_strings))))
        ax.set_yticklabels(y_strings)

    if colorbar is not None:
        colorbar = ax.figure.colorbar(colormesh, ax=ax, cax=colorbar.ax)
    else:
        colorbar = ax.figure.colorbar(colormesh, ax=ax)

    if z_is_stringy:
        N = len(z_strings)
        f = (N-1)/N
        colorbar.set_ticks([(n+0.5)*f for n in range(N)])
        colorbar.set_ticklabels(z_strings)

    return ax, colorbar


def _clip_nan_from_shaped_data(
    x: np.ndarray, y: np.ndarray, z: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    def _on_rectilinear_grid_except_nan(x_data: np.ndarray, y_data: np.ndarray) -> bool:
        """
        check that data is on a rectilinear grid. e.g. all points are the same as the  first
        row and column with the exception of nans. Those represent points not yet measured.
        """
        x_row = x_data[:, 0:1]
        y_row = y_data[0:1, :]
        return (
            np.nanmax(np.abs(x_data - x_row)) == 0
            and np.nanmax(np.abs(y_data - y_row)) == 0
        )

    if _on_rectilinear_grid_except_nan(x, y):
        # clip any row or column where there are nans in the first row
        # or column. Since we fill from here we assume that means that no data
        # has been measured for this row
        x_to_plot, y_to_plot = x[:, 0], y[0, :]

        filter_x = ~np.isnan(x_to_plot)
        filter_y = ~np.isnan(y_to_plot)

        x_to_plot = x_to_plot[filter_x]
        y_to_plot = y_to_plot[filter_y]
        z_to_plot = z[filter_x, :]
        z_to_plot = z_to_plot[:, filter_y].transpose()
    else:
        # fallback to flattening the data and use the same path as
        # non shaped data after filtering the nans.
        # this is not ideal as we loose the shape data but
        # not clear how to do this better. Either return a ragged
        # array or clip all inner dims that have nans completely
        x = x.flatten()
        y = y.flatten()
        z = z.flatten()
        filter_nans = np.logical_and(~np.isnan(x), ~np.isnan(y))
        x_to_plot, y_to_plot, z_to_plot = reshape_2D_data(
            x[filter_nans], y[filter_nans], z[filter_nans]
        )
    return x_to_plot, y_to_plot, z_to_plot


def _scale_formatter(tick_value: float, pos: int, factor: float) -> str:
    """
    Function for matplotlib.ticker.FuncFormatter that scales the tick values
    according to the given `scale` value.
    """
    return f"{tick_value*factor:g}"


def _make_rescaled_ticks_and_units(
    data_dict: DSPlotData,
) -> Tuple[matplotlib.ticker.FuncFormatter, str]:
    """
    Create a ticks formatter and a new label for the data that is to be used
    on the axes where the data is plotted.

    For example, if values of data are all "nano" in units of volts "V",
    then the plot might be more readable if the tick formatter would show
    values like "1" instead of "0.000000001" while the units in the axis label
    are changed from "V" to "nV" ('n' is for 'nano').

    The units for which unit prefixes are added can be found in
    `qcodes.utils.plotting._UNITS_FOR_RESCALING`. For all other units
    an exponential scaling factor is added to the label i.e.
    `(10^3 x e^2/hbar)`.

    Args:
        data_dict: A dictionary of the following structure
            {
                'data': <1D numpy array of points>,
                'name': <name of the parameter>,
                'label': <label of the parameter or ''>,
                'unit': <unit of the parameter or ''>
            }

    Returns:
        A tuple with the ticks formatter (matlplotlib.ticker.FuncFormatter) and
        the new label.
    """
    unit = data_dict['unit']

    maxval = np.nanmax(np.abs(data_dict['data']))
    prefix, selected_scale = find_scale_and_prefix(maxval, unit)

    new_unit = prefix + unit
    label = _get_label_of_data(data_dict)
    new_label = _make_axis_label(label, new_unit)

    scale_factor = 10**(-selected_scale)
    ticks_formatter = FuncFormatter(
        partial(_scale_formatter, factor=scale_factor))

    return ticks_formatter, new_label


def _rescale_ticks_and_units(
    ax: matplotlib.axes.Axes,
    data: Sequence[DSPlotData],
    cax: matplotlib.colorbar.Colorbar = None,
) -> None:
    """
    Rescale ticks and units for the provided axes as described in
    :func:`~_make_rescaled_ticks_and_units`
    """
    # for x axis
    if not _is_string_valued_array(data[0]['data']):
        x_ticks_formatter, new_x_label = \
            _make_rescaled_ticks_and_units(data[0])
        ax.xaxis.set_major_formatter(x_ticks_formatter)
        ax.set_xlabel(new_x_label)

    # for y axis
    if not _is_string_valued_array(data[1]['data']):
        y_ticks_formatter, new_y_label = \
            _make_rescaled_ticks_and_units(data[1])
        ax.yaxis.set_major_formatter(y_ticks_formatter)
        ax.set_ylabel(new_y_label)

    # for z aka colorbar axis
    if cax is not None and len(data) > 2:
        if not _is_string_valued_array(data[2]['data']):
            z_ticks_formatter, new_z_label = \
                _make_rescaled_ticks_and_units(data[2])
            cax.set_label(new_z_label)
            cax.formatter = z_ticks_formatter
            cax.update_ticks()


def _is_string_valued_array(values: np.ndarray) -> bool:
    """
    Check if the given 1D numpy array contains categorical data, or, in other
    words, if it is string-valued.

    Args:
        values: A 1D numpy array of values

    Returns:
        True, if the array contains string; False otherwise
    """
    return isinstance(values[0], str)
